# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

This file contains utility functions.

"""

import math
import numpy as np

try:
    import cupy as cp
except:
    pass

from PIL import Image


def get8bit(img):
    """Returns 8 bit representation of amplitude and phase of field.

    Returns a tuple of amplitude and phase, both real 2D numpy arrays of type
    uint8. Amplitude is scaled between 0 and 255, phase is wrapped and mapped
    to between 0 and 255, with 0 = 0 radians and 255 = 2pi radians.

    Parameters:
          img   : numpy.ndarray
                  2D numpy array, complex or real

    Returns:
          tuple of (ndarray, ndarray) : 8 bit amplitude and phase maps
    """

    if np.iscomplexobj(img):
        amp = np.abs(img).astype("float")
    else:
        amp = img
    amp = amp - np.min(amp)

    if np.max(amp) != 0:
        amp = amp / np.max(amp) * 255

    amp = amp.astype("uint8")

    phase = np.angle(img).astype("float")
    phase = phase % (2 * math.pi)
    phase = phase / (2 * math.pi) * 255
    phase = phase.astype("uint8")

    return amp, phase


def get16bit(img):
    """Returns 16 bit representation of amplitude and phase of field.

    Returns a tuple of amplitude and phase, both real 2D numpy arrays of type
    uint16. Amplitude is scaled between 0 and 2^16 -1, phase is mapped to between
    0 and 2^16 - 1, with 0 = 0 radians and 2^16 - 1 = 2pi radians.

    Parameters:
          img   : numpy.ndarray
                  2D numpy array, complex or real

    Returns:
          tuple of (ndarray, ndarray) : 16 bit amplitude and phase maps


    """
    amp = np.abs(img).astype("double")
    amp = amp - np.min(amp)
    if np.max(amp) != 0:
        amp = amp / np.max(amp) * 65535
    amp = amp.astype("uint16")

    phase = np.angle(img).astype("double")
    phase = phase % (2 * math.pi)
    phase = phase / (2 * math.pi) * 65535
    phase = phase.astype("uint16")

    return amp, phase


def save_phase_image(img, filename):
    """Saves phase as 16 bit tif. The phase is scaled so that 2pi = 65536.

    Parameters:
          img      : numpy.ndarray
                     2D numpy array, either complex field or real (phase map)
          filename : str
                     path to file to save to. If exists will be over-written.
    """

    if np.iscomplexobj(img):
        phase = np.angle(img).astype("double")
    else:
        phase = img.astype("double")
    phase = phase % (2 * math.pi)
    phase = ((phase / (2 * math.pi)) * 65536).astype("uint16")

    im = Image.fromarray(phase)
    im.save(filename)


def magnitude(img):
    """Returns magnitude of complex image.

    Parameters:
        img        : numpy.ndarray
                     complex image

    Returns:
        numpy.ndarray  : magnitude image
    """
    return np.abs(img) ** 2


def amplitude(img):
    """Returns amplitude of complex image. Deprecated, use `amp` instead.

    Parameters:
        img        : numpy.ndarray
                     complex image

    Returns:
         numpy.ndarray : amplitude image
    """
    return np.abs(img)


def amp(img):
    """Returns amplitude of complex image.

    Parameters:
        img        : numpy.ndarray
                     complex image

    Returns:
         numpy.ndarray : amplitude image
    """
    return np.abs(img)


def phase(img):
    """Returns phase of complex image, between 0 and 2pi.

    Parameters:
        img        : numpy.ndarray
                     complex image

    Returns:
         numpy.ndarray : phase map
    """
    return np.angle(img) % (2 * math.pi)


def load_image(filename):
    """
    Loads an image or stack of images from a file. Supports all formats supported by PIL.

    Parameters:
        filename    : str or Path
                      path to file

    Returns:
        numpy.ndarray : 2D, 3D or 4D numpy array representing image or stack

    """
    im = Image.open(filename)

    if im.n_frames > 1:
        h, w = np.shape(im)
        dt = np.array(im).dtype
        stack = np.zeros((im.n_frames, h, w), dtype=dt)
        for i in range(im.n_frames):
            im.seek(i)
            stack[i, :, :] = np.array(im)
        return stack
    else:
        return np.array(Image.open(filename))


def save_image(img, file, autoscale=True):
    """Saves an image stored as numpy array to an 8 bit image file.

    Parameters:
        img    : numpy.ndarray
                 image to save
        file   : str
                 filename to save to, type will be determined by extension.

    Optional Keyword Arguments:
        autoscale : boolean
                    if True (default), image is scaled to use full bit depth

    """
    if autoscale:
        img = Image.fromarray(get8bit(img)[0])
    else:
        img = Image.fromarray(img.astype("uint8"))
    img.save(file)


def save_image16(img, file, autoscale=True):
    """Saves an image stored as numpy array to an 16 bit image file.

    Parameters:
        img    : numpy.ndarray
                 image to save
        file   : str
                 filename to save to including extension, type will be
                 determined by extension and must support 16 bit images (e.g. tif)

    Optional Keyword Arguments:
        autoscale : boolean
                    if True (default), image is scaled to use full bit depth


    """
    if autoscale:
        img = Image.fromarray(get16bit(img)[0])
    else:
        img = Image.fromarray(img.astype("uint16"))
    img.save(file)


def save_amplitude_image8(img, filename):
    """Saves amplitude information as an 8 bit tif.

    Parameters:
        img    : numpy.ndarray
                 image to save
        file   : str
                 filename to load image from, including extension.
    """

    im = Image.fromarray(get8bit(img)[0])
    im.save(filename)


def save_amplitude_image16(img, filename):
    """Saves amplitude information as a 16 bit tif.

    Parameters:
        img    : numpy.ndarray
                 image to save
        file   : str
                 filename to load image from, including extension.
    """

    amp = amplitude(img)

    im = Image.fromarray(get16bit(img)[0])
    im.save(filename)


def extract_central(img, boxSize=None):
    """Extracts square of size boxSize*2 from centre of img. If boxSize is
    not specified, the largest possible square will be extracted.

    Parameters:
        img        : numpy.ndarray
                     complex or real image

    Keyword Arguments:
        boxSize    : int or None
                     size of square to be extracted

    Returns:
        numpy.ndarray : central square from image

    """
    w = np.shape(img)[0]
    h = np.shape(img)[1]

    cx = w / 2
    cy = h / 2
    if boxSize is not None:
        boxSemiSize = min(cx, cy, boxSize)
    else:
        boxSemiSize = min(cx, cy)

    imgOut = img[
        math.floor(cx - boxSemiSize) : math.floor(cx + boxSemiSize),
        math.ceil(cy - boxSemiSize) : math.ceil(cy + boxSemiSize),
    ]

    return imgOut


def invert(img):
    """Inverts an image, largest value becomes smallest and vice versa.

    Parameters:
        img        : numpy.ndarray
                     numpy array, input image

    Returns:
        numpy.ndarray : inverted image
    """

    return np.max(img) - img


def dimensions(inp):
    """Helper to obtain width and height in functions which accept multiple
    ways to send this information. The input may either be a single value,
    for a square image, a tuple of (h, w) or a 2D array.

    Parameters:
        inp        : int or (int, int) or ndarray

    Returns:
        tuple of (int, int), height and width
    """

    if type(inp) is np.ndarray:
        h, w = np.shape(inp)[0:2]
    elif type(inp) is tuple:
        h, w = inp
    else:
        w = inp
        h = inp

    return int(h), int(w)



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

This file contains functions related to estimating depth from shifted images.

This is experimental and does not form part of the documented API.

"""

import math
import numpy as np

try:
    import cupy as cp

    cudaAvailable = True
except:
    cudaAvailable = False
from PIL import Image
import cv2 as cv
import scipy
import time

import cv2 as cv

from pyholoscope.focus_stack import FocusStack
from pyholoscope.focusing_numba import propagator_numba
from pyholoscope.roi import Roi
from pyholoscope.utils import extract_central
import pyholoscope.general


def determine_shift(img1, img2, **kwargs):
    """Determines shift between two images by Normalised Cross Correlation (NCC). A sqaure template extracted
    from the centre of img2 is compared with a sqaure region extracted from the reference image img1. The size
    of the template (templateSize) must be less than the size of the reference (refSize). The maximum
    detectable shift is the (refSize - templateSize) / 2.

    Arguments:
        img1     : numpy.ndarray
                   reference image
        img2     : numpy.ndarray
                   image to be compared with reference

    Keyword Arguments:
        upsample : int
                   factor to scale images by prior to template matching to
                   allow for sub-pixel registration.
    """

    returnMax = kwargs.get("returnMax", False)
    upsample = kwargs.get("upsample", 1)

    refSize = np.shape(img1)[0] / 2
    templateSize = np.shape(img1)[0] / 4

    if (
        refSize < templateSize
        or min(np.shape(img1)) < refSize
        or min(np.shape(img2)) < refSize
    ):
        return -1
    else:
        template = extract_central(img2, templateSize).astype("float32")
        refIm = extract_central(img1, refSize).astype("float32")

        if upsample != 1:
            template = cv.resize(
                template,
                (np.shape(template)[0] * upsample, np.shape(template)[1] * upsample),
            )
            refIm = cv.resize(
                refIm, (np.shape(refIm)[0] * upsample, np.shape(refIm)[0] * upsample)
            )
        res = cv.matchTemplate(template, refIm, cv.TM_CCORR_NORMED)
        min_val, max_val, min_loc, max_loc = cv.minMaxLoc(res)
        shift = [
            (max_loc[0] - (refSize - templateSize) * upsample) / upsample,
            (max_loc[1] - (refSize - templateSize) * upsample) / upsample,
        ]

        if returnMax:
            return shift, max_val
        else:
            return shift


def calibrate_depth_from_shift(img1, img2, depth):
    pass



# -*- coding: utf-8 -*-
"""
Focus scoring functions and registry.
"""

import numpy as np
import scipy
import cv2 as cv

def get_focus_score_methods():
    """Returns a list of available focus scoring method names."""

    seen = set()
    result = {}

    # For backwards compatibility, there are some keys in the methods dict that point to the same function.
    # The deprecated names are always defined second, so we can loop through and only 
    # return the first occurance of each value.
    for key, value in methods.items():
        if value not in seen:
            result[key] = value
            seen.add(value)

    return list(result.keys())

def brenner(img):
    """Brenner focus metric (lower is better)."""
    (h, w) = np.shape(img)
    BrennX = np.zeros((h, w))
    BrennY = np.zeros((h, w))
    BrennX[0:-2, :] = img[2:, :] - img[0:-2,]
    BrennY[:, 0:-2] = img[:, 2:] - img[:, 0:-2]
    scoreMap = np.maximum(BrennY**2, BrennX**2)
    return -np.mean(scoreMap)

def peak(img):
    """Peak intensity focus metric (lower is better)."""
    return -np.max(img)

def sobel(img):
    """Sobel gradient energy focus metric (lower is better)."""
    filtX = np.array([[1, 0, -1], [2, 0, -2], [1, 0, -1]])
    filtY = np.array([[1, 2, 1], [0, 0, 0], [-1, -2, -1]])
    xSobel = scipy.signal.convolve2d(img, filtX)
    ySobel = scipy.signal.convolve2d(img, filtY)
    sobel = xSobel**2 + ySobel**2
    return -np.mean(sobel)

def sobel_variance(img):
    """Variance of Sobel energy focus metric (lower is better)."""
    filtX = np.array([[1, 0, -1], [2, 0, -2], [1, 0, -1]])
    filtY = np.array([[1, 2, 1], [0, 0, 0], [-1, -2, -1]])
    xSobel = scipy.signal.convolve2d(img, filtX)
    ySobel = scipy.signal.convolve2d(img, filtY)
    sobel = xSobel**2 + ySobel**2
    return -(np.std(sobel) ** 2)

def variance(img):
    """Intensity standard deviation focus metric (lower is better)."""
    return -np.std(img)

# https://doi.org/10.1016/j.optlaseng.2020.106195
def dark_focus(img):
    """DarkFocus metric from Optics & Laser Eng. 2020 (lower is better)."""
    kernelX = np.array([[-1, 0, 1]])
    kernelY = kernelX.transpose()
    gradX = cv.filter2D(img, -1, kernelX)
    gradY = cv.filter2D(img, -1, kernelY)
    mean, stDev = cv.meanStdDev(gradX**2 + gradY**2)
    return -(stDev[0, 0] ** 2)

def norm_var(img):
    """ Normalised variance focus metric (lower is better). """
    return -np.sum((img - np.mean(img))**2) / np.mean(img)
    
def sum_focus(img):
    """Sum of pixel intensities (lower is better)."""
    return np.sum(img)    

methods= {
    'brenner': brenner,
    'sobel': sobel,
    'dark_focus': dark_focus,
    'DarkFocus': dark_focus,           # Backwards compatibility
    'sobel_variance': sobel_variance,
    'SobelVariance': sobel_variance,   # Backwards compatibility
    'peak': peak,
    'norm_var': norm_var,
    'sum': sum_focus,
}

def focus_score(img, method):
    """Returns score of how 'in focus' an amplitude image is.
    Score is returned as a float, the lower the better the focus.

    Arguments:
        img          : numpy.ndarray
                       image to score, 2D real array
        method       : str or callable
                       scoring method name, or a callable accepting a 2D array
                       and returning a float (lower is better). Built-ins are:
                       'Brenner', 'Sobel', 'SobelVariance', 'Var',
                       'DarkFocus' or 'Peak'.
    Returns:
        float        : focus score
    """
    if isinstance(method, str):
        method = method.lower()
        if method not in methods:
            raise ValueError(f"Unknown focus scoring method: {method}")
        method = methods[method]
    elif not callable(method):
        raise ValueError("Method must be a string or a callable function.")   

    return method(img)




"""
PyHoloscope - Fast Holographic Microscopy for Python
"""

INLINE = 1
OFF_AXIS = 2

INLINE_MODE = 1  # deprecated, kept for backwards compatibility
OFFAXIS_MODE = 2  # deprecated, kept for backwards compatibility

from pyholoscope.general import *
from pyholoscope.off_axis import *
from pyholoscope.utils import *
from pyholoscope.prop_lut import *
from pyholoscope.focus_stack import *
from pyholoscope.roi import *
from pyholoscope.holo_class import *
from pyholoscope.focusing import *
from pyholoscope.phase_proc import *
from pyholoscope.windows import *


try:
    from pyholoscope.focusing_numba import *
except:
    pass
from pyholoscope.focus_shift import *



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

This file contains general functions.

"""

import warnings

import numpy as np
import cv2 as cv
import scipy as sp

from numba import jit, njit, prange


def pre_process(
    img,
    background=None,
    normalise=None,
    window=None,
    downsample=1.0,
    numba=False,
    precision="single",
):
    """Carries out steps required prior to refocusing.
    This includes background correction, normalisation and windowing and
    downsampling. It also coverts image to either type depending in specified
    precision (default is single precision).

    Parameters:
          img        : numpy.ndarray
                       raw hologram, 2D array, real or complex

    Keyword Arguments:
          background : numpy.ndaarray
                       backround hologram to be subtracted
                       2D real array (default = None)
          normalise  : numpy.ndarray
                       background hologram to be divided
                       2D real array (default = None)
          window     : numpy.ndarray
                       window to smooth edges. 2D real array. Will be
                       resized to match size of img if necessary.
                       (default = None)
          downsample : float
                       factor to downsample image by (default = 1)
          numba      : bool
                       flag to use numba for speed up (default = False)
          precision  : str
                       'single' or 'double' (default = 'single')
    Returns:
          numpy.ndarray  : pre-processed hologram, 2D array, real or complex

    """

    # Ensure the input hologram is a float or a complex float
    if np.iscomplexobj(img):
        if precision == "double":
            imType = "complex128"
        else:
            imType = "complex64"
    else:
        if precision == "double":
            imType = "float64"
        else:
            imType = "float32"

    if img.dtype != imType:
        img = img.astype(imType)

    if background is not None and not np.iscomplexobj(img):
        if background.dtype != imType:
            background = background.astype(imType)

    if normalise is not None and not np.iscomplexobj(img):
        if normalise.dtype != imType:
            normalise = normalise.astype(imType)

    img = remove_background(img, background)
    img = normalise_hologram(img, normalise)
    img = apply_window(img, window)
    img = downsample_hologram(img, downsample)

    return img


def remove_background(img, background):
    """Removes the background from a hologram.

    Arguments:
            img        : numpy.ndarray
                            raw hologram, 2D array, real or complex
            background : numpy.ndarray
                            background hologram to be subtracted, same shape as img
    Returns:
            numpy.ndarray  : hologram with background removed, 2D array, real or complex
    """

    if background is not None and np.shape(background) == np.shape(img):
        if np.iscomplexobj(img):
            imgAmp = np.abs(img)
            imgPhase = np.angle(img)
            img = (imgAmp - np.sqrt(background)) * np.exp(1j * imgPhase)
        else:
            img = img - background

    return img


def normalise_hologram(img, ref_img):
    """Normalises a hologram by dividing it by a reference image.

    Arguments:
            img        : numpy.ndarray
                            raw hologram, 2D array, real or complex
            ref_img    : numpy.ndarray
                            reference image to be divided by, same shape as img
    Returns:
            numpy.ndarray  : normalised hologram, 2D array, real or complex
    """

    if ref_img is not None and np.shape(ref_img) == np.shape(img):
        if np.iscomplexobj(img):
            imgAmp = np.abs(img)
            imgPhase = np.angle(img)
            img = (imgAmp / np.sqrt(ref_img)) * np.exp(1j * imgPhase)
        else:
            img = img / ref_img

    return img


def apply_window(img, window, allow_resize=True):
    """Applies a window to an image. If the window is not the same size as the image, it will be resized to match.

    Arguments:
            img        : numpy.ndarray
                         image to apply window to, 2D array, real or complex
            window     : numpy.ndarray
                         window to apply, 2D real array
    Keyword Arguments:
            allow_resize : bool
                           if True, allows the window to be resized to match the image size (default = True). Otherwise,
                           if the window is not the same size as the image, the function will return the image unchanged.
    Returns:
            None
    """

    if window is not None:
        # If the window is the wrong size, reshape it to match hologram
        if np.shape(window) != np.shape(img):
            if allow_resize:
                warnings.warn("Window needed resizing, may effect processing speed.")
                window = cv.resize(window, (np.shape(img)[1], np.shape(img)[0]))
            else:
                return img

        if np.iscomplexobj(img):
            img.imag = img.imag * window
            img.real = img.real * window
            img[img == -0 + 0j] = 0j  # Otherwise phase angle looks weird when plotted
        else:
            img = img * window
    return img


def downsample_hologram(img, factor):
    """Downsamples a hologram by a factor of 'factor'.
    Arguments:
            img        : numpy.ndarray
                         raw hologram, 2D array, real or complex
            factor     : int
                         factor to downsample by, should be an integer (default = 1)
    Returns:
            numpy.ndarray  : downsampled hologram, 2D array, real or complex
    """

    if factor != 1 and not np.iscomplexobj(img):
        img = cv.resize(
            img,
            (
                int(np.shape(img)[1] / factor / 2) * 2,
                int(np.shape(img)[0] / factor / 2) * 2,
            ),
        )
    return img


def fourier_plane_display(img):
    """Returns a real, log-scale Fourier plane for display purposes.

    Arguments:
          img             : numpy.ndarray
                            2D numpy array, real or complex.
    Returns:
           numpy.ndarray  : 2D real array, log scale of the Fourier plane.

    """

    fourierPlane = np.log(np.abs(np.fft.fftshift(np.fft.fft2(img))))

    return fourierPlane


def fast_full_2D_fft(input, use_numba=True):
    """ "
    Computes the full 2D FFT of a real input array using the rfft2 function and reconstructs the full FFT
    from the real FFT output by exploiting the symmetry. May be faster than np/sp fft2 for large arrays.

    Arguments:
          input : numpy.ndarray
                  2D numpy array, real
    Keyword Arguments:
          use_numba : bool
                      If True, uses numba for speed up (default = True). If this is set to False, this function
                      is likely to be slower than np.fft.fft2 or sp.fft.fft2.
    Returns:
            numpy.ndarray : 2D complex numpy array, full FFT of the input.

    """

    M, N = input.shape

    rfft = sp.fft.rfft2(input)

    if use_numba:
        # pass
        out = reconstruct_full_fft2(rfft, M, N)
    else:
        # Copy the rfft2 output into the left half
        mid = N // 2
        out = np.zeros((M, N), dtype="complex128")
        out[:, : mid + 1] = rfft

        if N % 2 == 0:
            reflected = np.conj(np.flipud(np.fliplr(rfft[1:, 1:])))
            out[1:, mid:] = reflected
            out[0, mid + 1 :] = np.conj(rfft[0, mid - 1 : 0 : -1])
        else:
            reflected = np.conj(np.flipud(np.fliplr(rfft[1:, 1:])))
            out[1:, mid + 1 :] = reflected
            out[0, mid + 1 :] = np.conj(rfft[0, mid:0:-1])

    return out



"""
PyHoloscope - Fast Holographic Microscopy for Python

This file contains functions for generating windows
for use in holographic microscopy.

"""

import numpy as np
import math
from pyholoscope import dimensions


def circ_window(img_size, circle_radius, data_type="float32"):
    """Produces a circular or elipitcal mask on grid of img_size as a numpy array.

    Parameters:
        img_size      : int or (int, int)
                       size of output array. Provide a single int to generate a square
                       image of that size, otherwise provide (w,h) to produce a rectangular
                       image, or 2D numpy array in which case the size of the array
                       will be used.
        circle_radius : float or (float, float)
                       Pixel values inside this radius will be 1. Provide a tuple
                       of (x,y) to have different x and y radii.

    Keyword Arguments:
        data_type     : str
                       data type of returned array (default is 'float32')

    Returns:
        numpy.ndarray : 2D numpy array containing mask


    """

    circle_radius = dimensions(circle_radius)
    [xM, yM] = np.meshgrid(range(circle_radius[0] * 2), range(circle_radius[1] * 2))
    mask = ((yM - circle_radius[1]) / circle_radius[1]) ** 2 + (
        (xM - circle_radius[0]) / circle_radius[0]
    ) ** 2 < 1

    return mask.astype(data_type)


def circ_cosine_window(img_size, circle_radius, skin_thickness, data_type="float32"):
    """Produces a circular or elliptical cosine window mask on grid of img_size as a numpy array.

    Parameters:
        img_size      :  int or (int, int)
                        size of output array. Provide a single int to generate a square
                       array of that size, otherwise provide (w,h) to produce a rectangular
                       array, or 2D numpy array in which case the size of the array
                       will be used.
        circle_radius :  float or (float, float)
                        Pixel values inside this radius will be 1. Provide a tuple
                        of (x,y) to have different x and y radii.
        skin_thickness : float
                        size of smoothed area inside circle/ellipse

    Keyword Arguments:
        data_type      : str
                        data type of returned array (default is 'float32')

    Returns:
        numpy.ndarray : 2D numpy array containing mask


    """
    h, w = dimensions(img_size)

    circle_radius = dimensions(circle_radius)
    skin_thickness = dimensions(skin_thickness)

    xM, yM = np.meshgrid(range(w), range(h))
    xMc = xM - w / 2
    yMc = yM - h / 2

    dist = np.sqrt(xMc**2 + yMc**2)
    xMc[xMc == 0] = 0.001
    theta = np.arctan(yMc / xMc)

    a = circle_radius[1]
    b = circle_radius[0]
    a2 = circle_radius[1] - skin_thickness[1]
    b2 = circle_radius[0] - skin_thickness[0]

    # Equations for radius of an ellipse at a given theta
    outerRadius = a * b / np.sqrt((a * np.cos(theta)) ** 2 + (b * np.sin(theta)) ** 2)

    innerRadius = (
        a2 * b2 / np.sqrt((a2 * np.cos(theta)) ** 2 + (b2 * np.sin(theta)) ** 2)
    )

    weight = (dist - innerRadius) / np.sqrt(
        (np.cos(theta) * skin_thickness[0]) ** 2
        + (np.sin(theta) * skin_thickness[1]) ** 2
    )

    # Smooth part of mask
    mask = np.cos(math.pi / 2 * (weight)) ** 2

    mask[weight > 1] = 0
    mask[weight < 0] = 1

    # Centre point will be NaN due to sqrt of 0
    mask[np.isnan(mask)] = 1

    return mask.astype(data_type)


def square_cosine_window(img_size, radius=None, skin_thickness=0, data_type="float32"):
    """Produces a square/rectangular cosine window mask on grid of img_size * img_size.
    Mask is 0 for pixels > radius and 1 for pixels < (radius -
    skin_thickness).  The intermediate region is a smooth squared cosine function.

    Arguments:
        img_size       : int or (int, int)
                        size of output array. Provide a single int to generate a square
                       array of that size, otherwise provide (w,h) to produce a rectangular
                       array, or 2D numpy array in which case the size of the array
                       will be used.

    Keyword Arguments:
        radius        :  float or (float, float) or None
                        Pixel values inside this radius will be 1. Provide a tuple
                        of (x,y) to have different x and y radii. If None (default), the radius will match
                        the size of the image.
        skin_thickness : float or None
                        size of smoothed area inside circle/ellipse. Default is 0.
        data_type      : str
                        data type of returned array (default is 'float32')

    Returns:
        numpy.ndarray : 2D numpy array containing mask


    """

    h, w = dimensions(img_size)

    if radius is None:
        radiusX = w / 2
        radiusY = h / 2
    elif type(radius) is tuple:
        radiusX, radiusY = radius
    else:
        radiusX = radius
        radiusY = radius

    innerRadX = radiusX - skin_thickness
    innerRadY = radiusY - skin_thickness

    xCentre = int(w / 2)
    yCentre = int(h / 2)

    yR = np.arange(h)
    xR = np.arange(w)

    row = np.cos(math.pi / (2 * skin_thickness) * (np.abs(xR - w / 2) - innerRadX)) ** 2
    row[np.abs(xR - xCentre) < innerRadX] = 1
    row[np.abs(xR - xCentre) > innerRadX + skin_thickness] = 0

    col = np.transpose(
        np.atleast_2d(
            np.cos(math.pi / (2 * skin_thickness) * (np.abs(yR - h / 2) - innerRadY))
            ** 2
        )
    )
    col[np.abs(yR - yCentre) < innerRadY] = 1
    col[np.abs(yR - yCentre) > innerRadY + skin_thickness] = 0

    maskH = np.tile(col, (1, w))
    maskV = np.tile(row, (h, 1))

    mask = maskH * maskV

    return mask.astype(data_type)



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

This file contains functions related to numerical refocusing.

"""

import math
import numpy as np

try:
    import cupy as cp

    cuda_available = True
except:
    cuda_available = False

try:
    from pyholoscope.focusing_numba import propagator_numba

    numba_available = True
except:
    numba_available = False


from PIL import Image
import scipy

from pyholoscope.focus_scores import *
from pyholoscope.focus_stack import FocusStack
from pyholoscope.focusing_numba import propagator_numba
from pyholoscope.roi import Roi
import pyholoscope.general
from pyholoscope.utils import dimensions
from pyholoscope.propagator import Propagator


def propagator(
    grid_size,
    wavelength,
    pixel_size,
    depth,
    geometry="plane",
    precision="single",
    use_numba=True,
):
    """Creates Fourier domain propagator for angular spectrum method.
    Returns the propagator as an instance of Propagator. Generation is sped up
    by only calculating top left quadrant and then duplicating
    (with flips) to create the other quadrants.

    Wavelength, pixel_size and depth are all specified in the same units.

    Arguments:
        grid_size   : float or (float, float)
                     size of square image (in pixels) to refocus, or
                     if tuple of (float, float), size of rectangular image as
                     height x width
        wavelength : float
                     wavelength of light
        pixel_size  : float
                     physical size of pixels
        depth      : float
                     refocus distance

    Keyword Arguments:
        geometry   : str
                     'plane' (default) or 'point'
        precision  : str
                     numerical precision of output, 'single' (default)
                     or 'double'
        use_numba   : boolean
                     if True, uses Numba version of functions (default is False)

    Returns:
        numpy.ndarray : 2D complex array, propagator

    """

    if precision == "double":
        data_type = "complex128"
    elif precision == "single":
        data_type = "complex64"
    else:
        raise Exception(f"Invalid precision {precision}, must be 'single' or 'double'.")

    grid_height, grid_width = dimensions(grid_size)
    if numba_available and use_numba:
        prop = propagator_numba(
            (grid_height, grid_width),
            wavelength,
            pixel_size,
            depth,
            geometry,
        )
        return Propagator(
            propagator=prop,
            wavelength=wavelength,
            pixel_size=pixel_size,
            depth=depth,
            geometry=geometry,
        )

    grid_height, grid_width = dimensions(grid_size)

    centre_x = int(grid_width // 2)
    centre_y = int(grid_height // 2)

    # Physical size of hologram in real units
    width = grid_width * pixel_size
    height = grid_height * pixel_size

    # Grid points to generate one quadrant of propagator on
    (xM, yM) = np.meshgrid(range(centre_x + 1), range(centre_y + 1))

    # Bins size of FFT (i.e. pixel size of FFT in inverse distance)
    delta0x = 1 / width
    delta0y = 1 / height

    # Generate one quadrant of the propagator
    if geometry == "point":
        u = delta0x * xM
        v = delta0y * yM
        prop_corner = np.exp(1j * math.pi * wavelength * depth * (u**2 + v**2))

    elif geometry == "plane":
        alpha = wavelength * xM / width
        beta = wavelength * yM / height
        prop_corner = np.exp(
            (-1j * 2 * math.pi * depth * np.sqrt(1 - alpha**2 - beta**2) / wavelength)
        )
        prop_corner[alpha**2 + beta**2 > 1] = 0

    else:
        raise Exception("Invalid geometry.")

    # Array to hold full propagator
    prop = np.zeros((grid_height, grid_width), dtype=data_type)

    # Duplicate the top left quadrant into the other three quadrants as
    # this is quicker then explicitly calculating the values
    prop[: centre_y + 1, : centre_x + 1] = prop_corner  # top left
    prop[: centre_y + 1, centre_x + grid_width % 2 :] = np.flip(
        prop_corner[:, 1:], 1
    )  # top right
    prop[centre_y + grid_height % 2 :, :] = np.flip(
        prop[1 : centre_y + 1, :], 0
    )  # bottom half

    prop_obj = Propagator(
        propagator=prop,
        wavelength=wavelength,
        pixel_size=pixel_size,
        depth=depth,
        geometry=geometry,
    )
    return prop_obj


def refocus(img, propagator, **kwargs):
    """Refocuses a hologram using the angular spectrum method.
    Takes a hologram 'hologram' wich may be a real or
    complex 2D numpy array (with any pre-processing
    such as background removal already performed) and a pre-computed 'propagator'
    which can be generated using the function 'propagator()'.

    Arguments:
        img           : ndarray
                        2D numpy array, raw hologram.
        propagator    : instance of Propagator
                        propagator object, containing the propagator as a
                        2D numpy array, as returned from propagator().

    Keyword Arguments:
        background    : numpy.ndarray or None
                        background image to subtract prior to refocus, same shape as img (default is None)
        normalise     : numpy.ndarray or None
                        normalisation image to divide by prior to refocus, same shape as img (default is None)
        window        : numpy.ndarray or None
                        spatial window to apply to image prior to refocus, same shape as img (default is None)
        downsample    : int or None
                        downsample factor, if not None, the image will be downsampled by this factor prior to refocus (default is None)
        fourier_domain : boolean
                        if True then img is assumed to be already the
                        FFT of the hologram, useful for speed when performing
                        multiple refocusing of the same hologram. (default is
                        False)
        cuda          : boolean
                        if True (default) GPU will be used if available.
        others        : pass any keyword arguments from pre_process() to
                        apply this pre-processing prior to refocusing

    Returns:
        numpy.ndarray : 2D complex array, refocused image
    """

    img_is_fourier = kwargs.pop("fourier_domain", False)
    cuda = kwargs.pop("cuda", True)

    # If we were sent the propagator on the CPU, push to GPU now
    if (
        cuda is True
        and cuda_available is True
        and type(propagator.propagator) is np.ndarray
    ):
        prop = cp.array(propagator.propagator)
    else:
        prop = propagator.propagator

    # If we have been sent the FFT of image, used when repeatedly calling refocus
    # (for example when finding best focus) we don't need to do FFT or shift for speed
    if img_is_fourier:
        if cuda is True and cuda_available is True:
            if type(img) is np.ndarray:
                img = cp.array(img)
            return cp.asnumpy(cp.fft.ifft2(img * prop))
        else:
            return scipy.fft.ifft2(img * prop)

    else:  # If we are sent the spatial domain image
        c_hologram = pyholoscope.pre_process(img, **kwargs)
        assert np.shape(c_hologram) == propagator.shape, (
            f"Propagator is the wrong size {propagator.shape} compared with hologram {np.shape(c_hologram)}."
        )

        if cuda is True and cuda_available is True:
            if type(c_hologram) is np.ndarray:
                c_hologram = cp.array(c_hologram)
            return cp.asnumpy(cp.fft.ifft2(cp.fft.fft2(c_hologram) * prop))
        else:
            return scipy.fft.ifft2(scipy.fft.fft2(c_hologram) * prop)


def correct_curvature(img, wavelength, pixel_size, source_distance):
    """Corrects for spherical wavefront curvature in hologram, assuming a point source.

    Arguments:
        img             : ndarray
                          2D numpy array, raw hologram.
        wavelength     : float
                          wavelength of light source
        pixel_size      : float
                          real pixel size of hologram
        source_distance: float
                          distance from point source to camera plane

    Returns:
        numpy.ndarray : 2D complex array, corrected hologram

    """

    (h, w) = np.shape(img)
    (xM, yM) = np.meshgrid(range(w), range(h))

    centre_x = int(w // 2)
    centre_y = int(h // 2)

    # Real space coordinates of each pixel relative to centre of image
    x = pixel_size * (xM - centre_x)
    y = pixel_size * (yM - centre_y)

    # Distance from point source to each pixel
    # r = np.sqrt(x**2 + y**2 + source_distance ** 2)
    phase = (x**2 + y**2) / (2 * source_distance)

    correction = np.exp(-1j * 2 * math.pi * phase / wavelength)

    return img * correction


def propagator_size_for_roi(img_shape, roi=None, margin=None):
    """Returns required propagator size for ROI + margin refocusing.

    Arguments:
        img_shape : tuple or ndarray
                    shape of full image, or image itself
        roi       : Roi or None
                    region of interest used for scoring
        margin    : int or None
                    margin around the ROI used for refocusing

    Returns:
        tuple (int, int) : (height, width) of propagator needed
    """

    h, w = dimensions(img_shape)

    if margin is None or roi is None:
        return (h, w)

    refocus_roi = Roi(
        roi.x - margin,
        roi.y - margin,
        roi.width + margin * 2,
        roi.height + margin * 2,
    )
    refocus_roi.constrain(0, 0, w, h)
    return (refocus_roi.height, refocus_roi.width)


def refocus_and_score(
    depth,
    img_fft,
    pixel_size,
    wavelength,
    method,
    score_roi=None,
    prop_lut=None,
    use_numba=False,
    use_cuda=False,
    precision="single",
):
    """Refocuses an image to specificed depth and returns focus score, used by
    findFocus.

    Parameters:
        depth     : float
                    depth to focus to
        img_fft   : ndarray, complex
                    FFT of hologram
        pixel_size: float
                    real pixel size of hologram
        wavelength: float
                    wavelength of light source
        method    : list of (str or callable) or str/callable,
                    scoring method (see focus_score for list). If a list is
                    provided, a list of scores will be returned, one for each
                    method, otherwise if a single method is provided as
                    a string or callable, then a single score will be returned.

    Keyword Arguments:
        score_roi  : ROI or None
                    region of image to apply scoring to (default is None, in
                    which case it applies to whole image.)
        prop_lut    : prop_lut or None
                    propgator look-up table (default is None, propagator is
                    calculate for each depth)
        use_numba  : boolean
                    if True, uses Numba version of functions (default is False)
        use_cuda   : boolean
                    if True, uses GPU where available
        precision : str
                    'single' (default) or 'double', precision of calculated propagator

    Returns:
        list of floats or float : either a list of scores or a single score.

    """

    # Whether we are using a look up table of propagators or calculating it each time
    if prop_lut is None:
        prop = propagator(
            img_fft,
            wavelength,
            pixel_size,
            depth,
            precision=precision,
            use_numba=use_numba,
        )

    else:
        prop = prop_lut.propagator(depth)

    # We are working with the FFT of the hologram for speed
    refocImg = np.abs(refocus(img_fft, prop, fourier_domain=True, cuda=use_cuda))

    # If we are running focus metric only on a ROI, extract the ROI
    if score_roi is not None:
        refocImg = score_roi.crop(refocImg)

    # If we have list of methods, we return a list of scores, otherwise
    # just a single score
    if isinstance(method, list):
        score = []
        for m in method:
            score.append(focus_score(refocImg, m))
    else:
        score = focus_score(refocImg, method)

    #print(depth, score)

    return score


def find_focus(img, wavelength, pixel_size, depth_range, method, **kwargs):
    """Determines the refocus depth which maximises the focus metric on image.

    To depth score using only a subset of the image, provide an instance of Roi as score_roi. Note that
    the entire image will still be refocused, i.e. this does not provide a speed improvement.

    To refocus only a small region of the image around the ROI (which is faster), provide a margin in margin,
    a region with this margin around the ROI will then be refocused. A pre-computed propagator LUT, an instance of prop_lut, can be
    provided in prop_lut. Note that if margin is specified, the propagator LUT must be of the correct size,
    i.e. the same size as the area to be refocused.

    To perform an initial coarse search to identify the region likely to have the best focus, provide the number of
    search regions to split the search range into in coarse_search_interval.

    Parameters:
        pixel_size  : float
                      real pixel size of hologram
        wavelength  : float
                      wavelength of light source
        depth_range : tuple of (float, float)
                      min and max depths to search between
        method      : str or callable
                      scoring method (see focus_score for list)

    Keyword Arguments:
        background : ndarray or None
                     background image (default is None)
        window     : ndarray or None
                     spatial window (default is None)
        score_roi  : ROI or None
                     region of image to apply scoring to (default is None, in
                     which case it applies to whole image.)
        margin     : int or None
                     if not none, only a region with this margin around the score_roi
                     will be refocused prior to scoring
        prop_lut   : prop_lut or None
                     propagator look up table (default is None)
        coarse_search_interval : int or None
                               number of intervals to divide depth range into
                               for initial search. Default is None, i.e. no
                               initial search.
        use_numba  : boolean
                     if True, uses Numba version of functions (default is False)
        use_cuda   : boolean
                     if True, uses GPU where available

    """
    background = kwargs.get("background", None)
    normalise = kwargs.get("normalise", None)
    window = kwargs.get("window", None)
    score_roi = kwargs.get("roi", None)
    margin = kwargs.get("margin", None)
    prop_lut = kwargs.get("prop_lut", None)
    coarse_search_interval = kwargs.get("coarse_search_interval", None)
    use_numba = kwargs.get("numba", False)
    use_cuda = kwargs.get("cuda", False)
    max_iter = kwargs.get("max_iter", 10)
    downsample = kwargs.get("downsample", 1)
    correct_curvature_bool = kwargs.get("correct_curvature", False)
    source_distance = kwargs.get("source_distance", None)
    

    c_hologram = pyholoscope.pre_process(
        img, background=background, normalise=normalise, window=window,
    )
    
    if correct_curvature_bool and source_distance is not None:
        c_hologram  = correct_curvature(
            c_hologram ,
            wavelength,
            pixel_size * downsample,
            source_distance,
        )


    # If a margin is specified, this means we only refocus the ROI plus a
    # margin around it for speed. Define the ROI here.
    if margin is not None and score_roi is not None:
        refocus_roi = Roi(
            score_roi.x - margin,
            score_roi.y - margin,
            score_roi.width + margin * 2,
            score_roi.height + margin * 2,
        )
        refocus_roi.constrain(0, 0, np.shape(img)[0], np.shape(img)[1])
        score_roi = Roi(margin, margin, score_roi.width, score_roi.height)
    else:
        refocus_roi = None

    # If we are only refocusing around a ROI, crop to the ROI + margn
    if refocus_roi is not None:
        cropped_img = refocus_roi.crop(c_hologram)
        score_roi.constrain(0, 0, np.shape(cropped_img)[0], np.shape(cropped_img)[1])
    else:
        cropped_img = c_hologram  # otherwise use whole image

    # Pre-compute the FFT of the hologram since we need this for every trial depth
    img_fft = scipy.fft.fft2(cropped_img)

    if coarse_search_interval is not None:
        startDepth = coarse_focus_search(
            img_fft,
            depth_range,
            coarse_search_interval,
            pixel_size,
            wavelength,
            method,
            score_roi,
            prop_lut,
        )
        interval_size = (depth_range[1] - depth_range[0]) / coarse_search_interval
        min_bound = max(depth_range[0], startDepth - interval_size)
        max_bound = min(depth_range[1], startDepth + interval_size)
        depth_range = [min_bound, max_bound]
    else:
        startDepth = (max(depth_range) - min(depth_range)) / 2

    # Find the depth using optimiser
    depth = scipy.optimize.minimize_scalar(
        refocus_and_score,
        method="bounded",
        bracket=depth_range,
        bounds=depth_range,
        options={"maxiter": max_iter},
        args=(
            img_fft,
            pixel_size,
            wavelength,
            method,
            score_roi,
            prop_lut,
            use_numba,
            use_cuda,
        ),
    )

    return depth.x


def coarse_focus_search(
    img_fft,
    depth_range,
    num_intervals,
    pixel_size,
    wavelength,
    method,
    score_roi,
    prop_lut,
):
    """An initial check for approximate location of good focus depths prior to a finer search.
    Used by findFocus, see this function for arguments.
    """
    search_depths = np.linspace(depth_range[0], depth_range[1], num_intervals)
    focus_score = np.zeros_like(search_depths)
    for idx, depth in enumerate(search_depths):
        focus_score[idx] = refocus_and_score(
            depth, img_fft, pixel_size, wavelength, method, score_roi, prop_lut
        )

    best_interval = np.argmin(focus_score)
    best_depth = search_depths[best_interval]

    return best_depth


def focus_score_curve(
    img, wavelength, pixel_size, depth_range, nPoints, method, **kwargs
):
    """Produces a plot of focus score against depth, mainly useful for debugging
    erroneous focusing.

    Arguments:
        img        : ndarray
                     2D numpy array, raw hologram
        wavelength : float
                     wavelength of light source
        pixel_size  : float
                     real pixel size of hologram
        depth_range : tuple of (float, float)
                     min and max depths to search between
        nPoints    : int
                     number of points to search between depth_range
        method     : func
                     scoring method (see focus_scores for list)

    Keyword Arguments:
        background : ndarray or None
                     background image (default is None)
        window     : ndarray or None
                     spatial window (default is None)
        score_roi  : ROI or None
                     region of image to apply scoring to (default is None, in
                     which case it applies to whole image.)
        margin     : int or None
                     if not none, only a region with this margin around the score_roi
                     will be refocused prior to scoring
        precision  : str
                     'single' (default) or 'double'
        use_numba  : boolean
                     if True, uses Numba version of functions (default is False)
        use_cuda   : boolean
                     if True, uses GPU where available
    """

    background = kwargs.get("background", None)
    normalise = kwargs.get("normalise", None)
    window = kwargs.get("window", None)
    score_roi = kwargs.get("roi", None)
    margin = kwargs.get("margin", None)
    precision = kwargs.get("precision", "single")

    c_hologram = pyholoscope.pre_process(
        img, background=background, normalise=normalise, window=window
    )

    if margin is not None and score_roi is not None:
        refocus_roi = Roi(
            score_roi.x - margin,
            score_roi.y - margin,
            score_roi.width + margin * 2,
            score_roi.height + margin * 2,
        )
        refocus_roi.constrain(0, 0, np.shape(img)[0], np.shape(img)[1])
        cropped_img = refocus_roi.crop(c_hologram)
    else:
        cropped_img = c_hologram

    # Do the forwards FFT just once for speed
    c_hologram_fft = scipy.fft.fft2(cropped_img)

    score = []
    depths = np.linspace(depth_range[0], depth_range[1], nPoints)
    for depth in depths:
        score.append(
            refocus_and_score(
                depth,
                c_hologram_fft,
                pixel_size,
                wavelength,
                method,
                score_roi,
                None,
                precision=precision,
            )
        )

    return score, depths


def refocus_stack(
    img,
    wavelength,
    pixel_size,
    depth_range,
    num_depths,
    **kwargs,
):
    """Generates a stack of images by refocusing a hologram to multiple
    depths.

    Parameters:
        wavelength : float
                     wavelength of light source
        pixel_size  : float
                     real pixel size of hologram
        depth_range : tuple of (float, float)
                     min and max depths of stack
        num_depths  : int
                     number of depths to refocus to within depth_range

    Keyword Arguments:
        background : ndarray or None
                     background hologram (default is None)
        window     : ndarray or None
                     spatial window (default is None)
        precision  : str
                     'single' (default) or 'double'
        use_numba  : boolean
                     if True, uses Numba version of functions (default is True)
        use_cuda   : boolean
                     if True, uses GPU where available

    Returns:
        instance of FocusStack
    """

    window = kwargs.get("window", None)
    use_numba = kwargs.get("use_numba", True)
    background = kwargs.get("background", None)
    precision = kwargs.get("precision", "single")
    geometry = kwargs.pop("geometry", "plane")

    depths = np.linspace(depth_range[0], depth_range[1], num_depths)

    # Apply pre-processing and then take 2D FFT
    c_hologram = pyholoscope.pre_process(img, **kwargs)
    c_hologram_fft = scipy.fft.fft2(c_hologram)

    # Tell refocus that we are providing the FFT
    refocus_kwargs = dict()
    refocus_kwargs["fourier_domain"] = True

    img = pyholoscope.pre_process(img, **kwargs)

    imgStack = FocusStack(c_hologram_fft, depth_range, num_depths)

    for idx, depth in enumerate(depths):
        prop = propagator(
            np.shape(img),
            wavelength,
            pixel_size,
            depth,
            precision=precision,
            use_numba=use_numba,
            geometry=geometry,
        )

        imgStack.add_idx(
            refocus(c_hologram_fft, prop, **refocus_kwargs),
            idx,
        )

    return imgStack



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

This file contains numba-optimised functions relaatd to numrical refocusing.

"""

import math

import numpy as np

import numba
from numba import jit, njit

from pyholoscope.utils import dimensions
from pyholoscope.propagator import Propagator


@jit(nopython=True)
def propagator_numba(
    grid_size, wavelength, pixel_size, depth, geometry="plane", precision="single"
):
    """Numba optimised version of propagator().
    Creates Fourier domain propagator for angular spectrum method.
    Returns the propagator as a complex 2D numpy array. Generation is sped up
    by only calculating top left quadrant and then duplicating
    (with flips) to create the other quadrants.

    Note that 'precision' is currently not implemented in the numba version.

    Parameters:
         gridSize   : (int, int)
                      size of image (in pixels) to refocus, height x width
         pixelSize  : float
                      physical size of pixels
         wavelength : float
                      in same units as pixelSize
         depth      : float
                      refocus depth in same units as pixelSize

    Keyword Arguments:
         geometry   : str
                      'plane' (default) or 'point'
         precision  : str
                      numerical precision of ouptut, 'single' (defualt)
                      or 'double' [NOT IMPLEMENTED]
    """    

    grid_width = int(grid_size[1])
    grid_height = int(grid_size[0])

    width = float(grid_width) * float(pixel_size)
    height = float(grid_height) * float(pixel_size)

    centre_x = int(grid_width // 2)
    centre_y = int(grid_height // 2)

    prop_corner = np.zeros((centre_y + 1, centre_x + 1), dtype=numba.complex64)
    prop = np.zeros((grid_height, grid_width), dtype=numba.complex64)

    delta0x = 1 / width
    delta0y = 1 / height

    if geometry == "point":
        fac = math.pi * wavelength * depth

        for x in range(centre_x + 1):
            uSq = (delta0x * x) ** 2

            for y in range(centre_y + 1):
                vSq = (delta0y * y) ** 2
                phase = fac * (uSq + vSq)

                # This is about as twice as fast as using np.exp(1j * phase)
                prop_corner.real[y, x] = math.cos(phase)
                prop_corner.imag[y, x] = math.sin(phase)

    elif geometry == "plane":
        fac = -2 * math.pi * depth / wavelength
        for x in range(centre_x + 1):
            alphaSq = (float(wavelength) * x * delta0x) ** 2

            for y in range(centre_y + 1):
                betaSq = (float(wavelength) * y * delta0y) ** 2
                if alphaSq + betaSq < 1:
                    phase = fac * np.sqrt(1 - alphaSq - betaSq)

                    # This is about as twice as fast as using np.exp(1j * phase)
                    prop_corner.real[y, x] = math.cos(phase)
                    prop_corner.imag[y, x] = math.sin(phase)
    else:
        raise Exception("Invalid geometry.")

    # Duplicate the top left quadrant into the other three quadrants as
    # this is quicker then explicitly calculating the values
    prop[: centre_y + 1, : centre_x + 1] = prop_corner  # top left
    prop[: centre_y + 1, (centre_x + grid_width % 2):] = prop_corner[:, 1:][:, ::-1]
    prop[centre_y + grid_height % 2 :, :] = prop[1 : centre_y + 1, :][::-1, :]
    return prop



## -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

The Holo Class provides an object-oriented interface to most of the
PyHoloscope functionality.
"""

import numpy as np
import time
import warnings

from pyholoscope.windows import (
    circ_cosine_window,
    circ_window,
    square_cosine_window,
)
from pyholoscope.off_axis import (
    off_axis_find_mod,
    off_axis_find_crop_radius,
    off_axis_demod,
)
from pyholoscope.focusing import (
    propagator,
    refocus,
    find_focus,
    refocus_stack,
    correct_curvature,
    propagator_size_for_roi,
)
from pyholoscope.general import pre_process
from pyholoscope.phase_proc import relative_phase
from pyholoscope.prop_lut import PropLUT
from pyholoscope.roi import Roi
from pyholoscope.utils import dimensions


# Check if cupy is available
try:
    import cupy as cp

    cuda_available = True
except:
    cuda_available = False

# Check if numba is available and if it runs without an error
try:
    import numba
    from pyholoscope.focusing_numba import propagator_numba

    test_prop = propagator_numba(
        (int(6), int(6)),
        float(1.0),
        float(1.0),
        float(1.0),
        geometry="plane",
        precision="single",
    )  # Run the JIT once for speed
    numba_available = True

except:
    numba_available = False


class Holo:
    # Processing pipeline
    INLINE = 1
    OFF_AXIS = 2
    modes = ["NONE", "INLINE", "OFF AXIS"]
    INLINE_MODE = 1  # deprecated, kept for backwards compatibility
    OFFAXIS_MODE = 2  # deprecated, kept for backwards compatibility

    # Off-axis crop mask types
    CUSTOM = 1
    CIRCLE = 2
    CIRCLE_COSINE = 3

    # For off-axis holography, these are generated if needed from the
    # background and normalisation holograms later
    background_field = None
    background_abs = None
    background_angle = None
    normalise_field = None
    normalise_abs = None
    normalise_angle = None
    __crop_window = None

    # Standard image type
    image_type = "float32"

    propagator = None
    propagator_lut = None
    auto_focus_propagator_lut = None

    def __init__(self, mode=None, wavelength=None, pixel_size=None, **kwargs):
        """Initialises an instance of the Holo class.

        Mode Parameters:

            mode:   enum
                        Processing mode, either pyholoscope.INLINE or pyholsocope.OFF_AXIS.

        Numerical Refocusing Parameters:
            refocus: bool
                        Flag to enable numerical refocusing (default = False) in OFF_AXIS mode. For INLINE mode,
                        refocusing is always enabled and this parameter is ignored.
            wavelength: float
                        Wavelength of light (m). Only needed if refocusing is required.
            pixel_size: float
                        Size of the pixels in the hologram (m). Only needed if refocusing is required.
            depth:  float
                        Refocus depth in same units as pixel_size and wavelength. Only needed if refocusing is required.
            source_distance: float or None
                        Distance from point source to image plane in same units as pixel_size and wavelength. Only required
                        if performing curvature correction. (default = None).
            correct_curvature: bool
                        Flag to enable curvature correction (default = False). If True, the hologram will be corrected
                        for spherical wavefront curvature before refocusing. Requires source_distance to be set.
            geometry: str
                        Geometry of the angular spectum propagator used for refocusing, 'plane' or 'point' (default = 'plane').
                        Only needed if refocusing is required.
            use_prop_lut : boolean
                           If True, use propagator LUT to refocus.

        Auto Focus Parameters:
            find_focus_depth_range: tuple
                        Depth range to use for finding the focus, (min_depth, max_depth) in same units as pixel_size and wavelength.
                        default = (0, 1).
            find_focus_method: str
                        Method to use for autofocus, (default = 'Brenner'). Only needed if autofocus is
                        used.
            find_focus_roi: Roi or None
                        Region of interest to use for finding the autofocus, (default = None). Only used for autofocus.
            find_focus_margin: int or None
                        Margin to use for autofocus, in pixels (default = None). Only used for autofocus.
            find_focus_coarse_search_interval: int or None
                        Interval to use for coarse search during autofocus, in m (default = None, in which case the coarse
                        search is not used).


        Backgound and Normalisation Parameters:
            background: numpy.ndarray  or None
                        Background hologram to be subtracted in inline holography, a 2D real array or None for no subtraction
                        (default = None). This will also be used to create the background field in off-axis holography if relative
                        phase is requested.
            normalise: numpy.ndarray  or None
                        Normalisation hologram to be divided by, 2D real array (default = None) during inline
                        or off-axis holography.
            relative_amplitude: bool
                        Flag to calculate relative amplitude in off-axis holography (default = False). If True, the background
                        hologram will be used to calculate the relative amplitude.
            relative_phase: bool
                        Flag to make the phase relative to the mean phase in the whole image or a specified ROI (default = False)
                        in off-axis holography. !!!!

        Windowing Parameters:
            auto_window: bool
                        Flag to automatically create a window for pre or post processing (default = False). If True, a window will be created
                        based on the image size and the specified window radius and thickness.
            post_window: bool
                        Flag to apply the window after refocusing (default = False). If True, the window will be applied after refocusing.
            window_shape: str
                        Shape of the automatically generated window, 'circle' or 'square' (default = 'square'). Only used if auto_window is True.
            window_radius: int or tuple
                        Radius of the automatically generated window, for 'circle' this is the radius, for 'square' this is half the side length.
                        For 'circle' provide an int, for 'square' either provide an int (resulting in a square window) or a tuple of (width, height).
                        (default = None). If None, the window radius will be set to half the image size.
            window_thickness: int
                        The number of pixels inside the window over which it transitions from opaque to transparent (default = 10).
                        Only used if auto_window is True.
            window: numpy.ndarray or None
                        A custom window as 2D real array. Will be resized to match size of img if necessary.
                        (default = None). If None, the window will be created automatically if auto_window is True.

        Off-Axis Demodulation Parameters:
            crop_centre: tuple or None
                        Centre of the crop in the off-axis demodulation, (x, y) in pixels (default = None). If None, the calib_off_axis()
                        method must be called to find the crop centre.
            crop_radius: int or None
                        Radius of the crop in the off-axis demodulation, in pixels (default = None). If None, the calib_off_axis()
                        method must be called to find the crop radius.
            crop_mask: numpy.ndarray or None
                        Shape of crop mask to use in the off-axis demodulation, pyholoscope.CIRCLE, pyholoscope.CIRCLE_COSINE, or
                        pyholsocope.CUSTOM. (default = None). If specified as pyholoscope.CUSTOM, the mask must be specified
                        using the custom_crop_window parameter. Otherwise, the crop mask will be a circle or a circle with a cosine
                        transition depending on the value of crop_mask and will be generated either when calib_off_axis() is called
                        or when process() is first called.
            custom_crop_window: numpy.ndarray or None
                        Custom crop window to use in the off-axis demodulation when crop_mask is set to pyholsocope.CUSTOM.
                        Provide a 2D real array (default = None).
            crop_window_skin_thickness: int
                        When crop_mask = pyholoscope.CIRCLE_COSINE, this parameter contols the number of pixels inside the
                        crop window over which it transitions from opaque to transparent (default = 10).
            return_fft: bool
                        Flag to return the FFT of the off-axis demodulated image (default = False).
                        If True, the process() method will be return a tuple of (demodulated image, fft)
            off_axis_real_fft: bool
                        Flag to use a real FFT for off-axis demodulation (default = False). If True, the real FFT will be used instead
                        of the complex FFT. This is faster but should only be used if the reference beam is not tilted in such a way that
                        the cross-term crosses the vetical axis in the FFT.
        Phase Processing Parameters:
            relative_phase: bool
                        Flag to make the phase relative to the mean phase in the whole image or a specified ROI (default = False).
            stable_phase: bool
                        Flag to make the phase stable, i.e. to remove the global phase from the image (default = False).
            stable_roi: pyholoscope.Roi or None
                        Region of interest to use for making the phase stable. In the output image
                        the mean phase in this region will be zero. (default = None). If None, the whole image will be used.

        Display Parameters:
                invert: bool
                        Flag to invert the image, i.e. largest value becomes smallest and vice versa (default = False).
            downsample: float
                        Factor to downsample the image by (default = 1). If > 1, the image will be downsampled by this factor.


        Back-end and Processing Parameters:
            numba:  bool
                        Flag to use numba for speed up (default = True). If False, numba optimised functions will not be used.
            cuda:   bool
                        Flag to use CUDA for GPU acceleration (default = True). If False, GPU functions will not be used.
            precision: str
                        Numerical precision of output, 'single' (default) or 'double'. This will determine the data type of the images
                        processed by the Holo class. If 'double', images will be processed as float64, otherwise as float32.
        """

        self.mode = mode
        self.wavelength = wavelength
        self.pixel_size = pixel_size
        self.oa_pixel_size = pixel_size

        # Numerical refocusing
        self.depth = kwargs.get("depth", 0)
        self.set_background(kwargs.get("background", None))
        self.set_normalise(kwargs.get("normalise", None))
        self.geometry = kwargs.get("geometry", "plane")
        self.use_prop_lut = kwargs.get("use_prop_lut", False)
        self.source_distance = kwargs.get("source_distance", None)
        self.correct_curvature = kwargs.get("correct_curvature", False)

        # Widowing
        self.auto_window = kwargs.get("auto_window", False)
        self.post_window = kwargs.get("post_window", False)
        self.window = kwargs.get("window", None)
        self.window_shape = kwargs.get("window_shape", "square")
        self.window_radius = kwargs.get("window_radius", None)
        self.window_thickness = kwargs.get("window_thickness", 10)

        # Autofocus
        self.find_focus_method = kwargs.get("find_focus_method", "Brenner")
        self.find_focus_roi = kwargs.get("find_focus_roi", None)
        self.find_focus_margin = kwargs.get("find_focus_margin", None)
        self.find_focus_coarse_search_interval = kwargs.get(
            "find_focus_coarse_search_interval", None
        )
        self.find_focus_depth_range = kwargs.get("find_focus_depth_range", (0, 1))

        # Off-axis demodulation
        self.crop_centre = kwargs.get("crop_centre", None)
        self.crop_radius = kwargs.get("crop_radius", None)
        self.return_fft = kwargs.get("return_fft", False)
        self.crop_mask = kwargs.get("crop_mask", None)
        self.custom_crop_window = kwargs.get("custom_crop_window", None)
        self.__crop_window_skin_thickness = kwargs.get("crop_window_skin_thickness", 10)
        self.off_axis_real_fft = kwargs.get("off_axis_real_fft", False)
        self.relative_amplitude = kwargs.get("relative_amplitude", False)

        # Phase
        self.relative_phase = kwargs.get("relative_phase", False)
        self.stable_phase = kwargs.get("stable_phase", False)
        self.stable_roi = kwargs.get("stable_roi", False)

        # Display
        self.invert = kwargs.get("invert", False)
        self.refocus = kwargs.get("refocus", False)
        self.downsample = kwargs.get("downsample", 1)

        # GPU and Numba
        self.cuda_available = cuda_available
        self.use_numba = kwargs.get("numba", True)
        self.cuda = kwargs.get("cuda", True)

        # Image data type
        self.set_precision(kwargs.get("precision", "single"))

    def __process_inline(self, img):
        """Process an inline hologram image, img, using the currently selected
        parameters.

        Argumnents:
            img: numpy.ndarray
                The hologram image to be processed as a 2D numpy array.
        Returns:
            img_out: numpy.ndarray
                The processed hologram image as a 2D numpy array.
        """

        # If we doing auto_window, and we either don't have
        # a window, or it is the wrong size, we make a new window
        self.update_auto_window(img)

        img_preprocessed = pre_process(
            img,
            downsample=self.downsample,
            window=self.window,
            background=self.background,
            normalise=self.normalise,
            precision=self.precision,            
        )

        if self.correct_curvature and self.source_distance is not None:
            img_preprocessed = correct_curvature(
                img_preprocessed,
                self.wavelength,
                self.pixel_size * self.downsample,
                self.source_distance,
            )

        if self.use_prop_lut:
            if self.propagator_lut is not None:
                propagator = self.propagator_lut.propagator(self.depth)

                if propagator is None:
                    raise "Requested refocus depth is outside of propagator LUT range."
            else:
                raise "Propagator LUT has not been generated."

        else:
            # If the propagator is not the correct one for the current parameters, regenerate it
            if (
                self.propagator is None
                or self.propagator.has_attributes(
                    wavelength=self.wavelength,
                    pixel_size=self.pixel_size,
                    depth=self.depth,
                )
                is False
            ):
                self.update_propagator(img)

            if np.shape(self.propagator.propagator) != np.shape(img_preprocessed):
                self.update_propagator(img_preprocessed)

            propagator = self.propagator

        # Numerical refocusing
        img_out = refocus(
            img_preprocessed, propagator, cuda=(self.cuda and cuda_available)
        )
        if img_out is None:
            warnings.warn("Output from refocusing was None.")
            return None

        # Post refocusing processing
        if self.post_window is True and self.window is not None:
            img_out = pre_process(img_out, window=self.window, precision=self.precision)

        if self.invert is True:
            img_out = np.max(img_out) - img_out

        return img_out

    def __process_off_axis(self, img):
        """Process an off-axis hologram image using the currently selected parameters."""

        assert self.crop_centre is not None, (
            "Off-Axis demodulation frequency not defined."
        )
        assert self.crop_radius is not None, "Off-Axis demodulation radius not defined."

        if self.crop_mask == self.CUSTOM and self.custom_crop_window is not None:
            self.__crop_window = self.custom_crop_window
        elif self.crop_mask == self.CIRCLE or self.crop_mask == self.CIRCLE_COSINE:
            if self.__crop_window is None:
                self.__create_off_axis_crop_window()

        # Off Axis Demodulation
        demod = off_axis_demod(
            img,
            self.crop_centre,
            self.crop_radius,
            mask=self.__crop_window,
            return_fft=self.return_fft,
            cuda=self.cuda,
            real_fft=self.off_axis_real_fft,
        )

        if demod is None:
            warnings.warn("Output from off-axis demodulation was None.")
            return None

        # If return_fft is True, off_axis_demod returns the demodulated image and the FFT as a tuple. If we
        # have been asked for the FFT we pull this out and return it, otherwise 'demod' is already just the demodulated image
        # and we continue
        if self.return_fft:
            return demod[1]

        # Relative phase means to subtract the phase from the background image
        if self.relative_phase:
            if self.background_field is not None:
                demod = relative_phase(demod, self.background_field)
            elif (
                self.background is not None
            ):  # Need to re-generate the background field
                self.__off_axis_background_field()
                demod = relative_phase(demod, self.background_field)

            else:
                warnings.warn(
                    "Relative phase requested but no background field available, call __off_axis_background_field() to create this first."
                )

        if demod is None:
            warnings.warn("Output from off-axis relative phase was None.")
            return None

        # If we have are doing auto_window, and we either don't have
        # a window, or it is the wrong size, we make a new window
        self.update_auto_window(img)

        # Off axis demodulation changes the pixel size,
        # so here we calculate the corrected pizel size
        if self.pixel_size is not None:
            self.oa_pixel_size = (
                self.pixel_size / float(np.shape(demod)[0]) * float(np.shape(img)[0])
            )

        # Apply background, normalisation, windowing, downsampling
        if self.relative_amplitude:
            background = self.background_abs
        else:
            background = None

        if self.normalise is not None:
            if self.normalise_abs is None or np.shape(self.normalise_abs) != np.shape(
                demod
            ):
                self.__off_axis_normalise_field()

        if self.relative_amplitude and self.background is not None:
            if self.background_abs is None or np.shape(self.background_abs) != np.shape(
                demod
            ):
                self.__off_axis_background_field()

        demod = pre_process(
            demod,
            downsample=self.downsample,
            window=self.window,
            background=background,
            normalise=self.normalise_abs,
            precision=self.precision,
        )

        # Numerical refocusing
        if self.refocus is True:
            # Check the propagator is valid, otherwise recreate it
            if self.propagator is None or not self.propagator.has_attributes(
                depth=self.depth,
                wavelength=self.wavelength,
                pixel_size=self.oa_pixel_size,
            ):
                self.update_propagator(demod)

            # Refocus
            demod = refocus(demod, self.propagator, cuda=(self.cuda and cuda_available))

            if demod is None:
                warnings.warn("Output from off-axis refocusing was None.")
                return None

        # Post refocusing processing
        if demod is not None:
            if (
                self.post_window is True
                and self.auto_window is True
                and self.window is not None
            ):
                demod = pre_process(demod, window=self.window, precision=self.precision)

            if self.invert is True:
                demod = np.max(demod) - demod

        if demod is None:
            warnings.warn("Output from off-axis processing was None.")
            return None

        return demod

    def __apply_window(self, img):
        """Applies the current window to a hologram 'img'."""
        img = pre_process(img, window=self.window)
        return img

    ###############################################################################
    ####################################### API ###################################

    def process(self, img):
        """Process a hologram using the currently selected parameters.
        Calls _process_inline or __process_off_axis depending  on mode.
        """

        # If we are refocusing we must have a wavelength, pixel size and depth specified
        if self.mode == self.INLINE or (
            self.mode == self.OFF_AXIS and self.refocus == True
        ):
            assert self.pixel_size is not None, "Pixel size not specified."
            assert self.wavelength is not None, "Wavelength not specified."
            assert self.depth is not None, "Refocus depth not specified."

        if img is None:
            warnings.warn("Image provided to process was None, output will be None.")
            return None

        assert img.ndim == 2, "Input must be a 2D numpy array."

        if self.mode == self.INLINE_MODE or self.mode == self.INLINE:
            return self.__process_inline(img)
        elif self.mode == self.OFFAXIS_MODE or self.mode == self.OFF_AXIS:
            return self.__process_off_axis(img)
        else:
            raise Exception("Invalid processing mode.")

    def set_mode(self, mode):
        self.mode = mode

    def set_refocus(self, refocus):
        self.refocus = refocus

    ############### PHYSICAL PARAMETERS ##########################################

    def set_wavelength(self, wavelength):
        """Set the wavelength of the hologram"""
        self.wavelength = wavelength

    def set_pixel_size(self, pixel_size):
        """Set the size of pixels in the raw hologram"""
        self.pixel_size = pixel_size

    ########### BACKGROUND AND FLAT-FIELDING #####################################

    def set_background(self, background):
        """Set the background hologram. Use None to remove background."""
        self.clear_background()
        if background is not None:
            self.background = background.astype(self.image_type)
        else:
            self.background = None

    def set_normalise(self, normalise):
        """Set the normalisation hologram. Use None to remove normalisation."""
        self.clear_normalise()
        if normalise is not None:
            self.normalise = normalise.astype(self.image_type)

    def clear_background(self):
        """Remove existing background hologram."""
        self.background = None
        self.background_field = None
        self.background_abs = None
        self.background_angle = None

    def clear_normalise(self):
        """Remove existing normalisation hologram."""
        self.normalise = None
        self.normalise_field = None
        self.normalise_abs = None
        self.normaliseAngle = None

    def set_relative_amplitude(self, boolean):
        """Sets whether or not to calculate relative amplitude in off-axis holography."""
        assert boolean == True or boolean == False, (
            "Argument of set_relative_amplitude must be True or False"
        )
        self.relative_amplitude = boolean

    ############################# WINDOW #####################################

    def __create_window(self, img_size, radius, skin_thickness, shape="square"):
        """Creates and stores the window used for pre or post processing.

        Arguments:
            img_size       :  the size of the window array, must be the same as the hologram it will be
                             applied to. Either provide a 2D numpy array, in which case the window will
                             be created to match the size of this, provide an int, in which case the window
                             will be a square of this size or a tuple of (height, width).
            radius        :  the size of the  transparent part of the window, for 'circle' this is the
                             radius, for 'square' this is half the side length. For 'circle' provide
                             an int, for 'square' either provide an int (resulting in a square window)
                             or a tuple of (height, width) for rectangular window.
            skin_thickness :  The number of pixels inside the window over which it transitions from
                             opaque to transparent.

        Keyword Arguments:
            shape         :  [Optional] window shape, 'circle' or 'square' (defualt).
        """

        if shape == "circle":
            self.window = circ_cosine_window(
                img_size, radius, skin_thickness, data_type=self.image_type
            )
        elif shape == "square":
            self.window = square_cosine_window(
                img_size, radius, skin_thickness, data_type=self.image_type
            )

    def set_window(self, window):
        """Sets the window to a pre-generated 'window', a 2D numpy array."""
        self.clear_window()
        if window is not None:
            self.window = window.astype(self.image_type)

    def clear_window(self):
        """Removes existing window, equivalent to set_window(None)"""
        self.window = None

    def set_auto_window(self, auto_window):
        """Sets whether or not use auto create a window (boolean)."""
        assert auto_window == True or auto_window == False, (
            "set_auto_window must be True or False"
        )
        self.auto_window = auto_window

    def set_post_window(self, post_window):
        """Sets whether or not to apply the window after refocusing (boolean)."""
        assert post_window == True or post_window == False, (
            "set_post_window must be True or False"
        )
        self.post_window = post_window

    def update_auto_window(self, img):
        """Create or re-create the automatic window using current parameters.
        Provide an 'img', a 2D numpy array of the same size as the image to
        be processed.
        """

        im_height = np.shape(img)[0]
        im_width = np.shape(img)[1]

        if self.auto_window == True:
            if self.window is None:
                need_to_regenerate_window = True
            elif (
                np.shape(self.window)[0] != np.shape(img)[0] / self.downsample
                or np.shape(self.window)[1] != np.shape(img)[1] / self.downsample
            ):
                need_to_regenerate_window = True
            else:
                need_to_regenerate_window = False

            if need_to_regenerate_window:
                if self.window_radius is None:
                    window_radius_x = int(im_width / 2)
                    window_radius_y = int(im_height / 2)
                else:
                    window_radius_x, window_radius_y = dimensions(self.window_radius)

                self.__create_window(
                    (int(im_height / self.downsample), int(im_width / self.downsample)),
                    (
                        int(window_radius_x / self.downsample),
                        int(window_radius_y / self.downsample),
                    ),
                    self.window_thickness / self.downsample,
                    shape=self.window_shape,
                )

    def set_window_shape(self, window_shape):
        """Sets the auto window shape, 'cicle' or 'square'."""
        if window_shape == "circle" or window_shape == "square":
            self.window_shape = window_shape
        else:
            raise Exception("Invalid window shape.")

    def set_window_radius(self, window_radius):
        """Sets the auto window radius."""
        self.window_radius = window_radius

    def set_window_thickness(self, window_thickness):
        """Sets the auto window edge thickness."""
        self.window_thickness = window_thickness

    ################## OFF AXIS DEMODULTION ######################################

    def set_off_axis_mod(self, crop_centre, crop_radius):
        """Sets the location of the frequency domain position of the OA modulation.

        Arguments:
            crop_centre  : tuple of (x,y)
            crop_radius  : radius
        """
        self.crop_centre = crop_centre
        self.crop_radius = crop_radius

    def set_crop_centre(self, centre):
        """Set the location of the modulation frequency in frequency domain.
        'centre' is a tuple of (x,y).
        """
        self.crop_centre = dimensions(centre)

    def set_crop_radius(self, radius):
        """Set the size of the region to extract in frequency domain to demodulate."""
        self.crop_radius = dimensions(radius)

    def set_return_FFT(self, return_fft):
        """Sets whether the FFT, rather than the demodualted image, is returned in OAH.
        Set True to obtain FFT, False to obtain image.
        """
        self.return_fft = return_fft

    def auto_find_off_axis_mod(self, maskFraction=0.1):
        """Detect the modulation location in frequency domain. maskFraction
        is the size of a mask applied to the centre of the FFT to prevent
        the d.c. from being detected.
        """
        if self.background is not None:
            self.crop_centre = off_axis_find_mod(
                self.background, maskFraction=0.1, real_fft=self.off_axis_real_fft
            )
            self.crop_radius = off_axis_find_crop_radius(
                self.background, maskFraction=0.1, real_fft=self.off_axis_real_fft
            )

    def calib_off_axis(self, hologram=None, mask_fraction=0.1):
        """Detect the modulation location in frequency domain using the
        background or a provided hologram.
        """

        if hologram is None:
            hologram = self.background

        if hologram is None:
            raise Exception(
                "Calib_off_axis requires a calibration image, either provided as an argument or from a previously set background."
            )

        self.crop_centre = off_axis_find_mod(
            hologram, mask_fraction=mask_fraction, real_fft=self.off_axis_real_fft
        )
        self.crop_radius = off_axis_find_crop_radius(
            hologram, mask_fraction=mask_fraction, real_fft=self.off_axis_real_fft
        )

        if self.background is not None:
            self.__off_axis_background_field()
        if self.normalise is not None:
            self.__off_axis_normalise_field()

        self.__create_off_axis_crop_window()

    def __create_off_axis_crop_window(self):
        """Create the crop window used for off-axis demodulation."""

        if self.crop_mask == self.CIRCLE:
            if isinstance(self.crop_radius, tuple):
                self.__crop_window = circ_window(
                    (self.crop_radius[0] * 2, self.crop_radius[1] * 2), self.crop_radius
                )
        elif self.crop_mask == self.CIRCLE_COSINE:
            if isinstance(self.crop_radius, tuple):
                self.__crop_window = circ_cosine_window(
                    (self.crop_radius[0] * 2, self.crop_radius[1] * 2),
                    self.crop_radius,
                    self.__crop_window_skin_thickness,
                )
        elif self.crop_mask == self.CUSTOM:
            self.__crop_window = self.custom_crop_window
        else:
            self.__crop_window = None

    def __off_axis_background_field(self):
        """Demodulate the background hologram."""
        assert self.background is not None, "Background hologram not provided."
        assert self.crop_centre is not None, "Demodulation centre not provided"
        assert self.crop_radius is not None, "Demodulation radius not provided."
        self.background_field = off_axis_demod(
            self.background,
            self.crop_centre,
            self.crop_radius,
            real_fft=self.off_axis_real_fft,
        )
        self.background_abs = np.abs(self.background_field)  # Store these now for speed
        self.backgroundPhase = np.angle(self.background_field)

    def __off_axis_normalise_field(self):
        """Demodulate the background hologram."""
        assert self.background is not None, "Background hologram not provided."
        assert self.crop_centre is not None, "Demodulation centre not provided"
        assert self.crop_radius is not None, "Demodulation radius not provided."
        self.normalise_field = off_axis_demod(
            self.normalise,
            self.crop_centre,
            self.crop_radius,
            real_fft=self.off_axis_real_fft,
        )
        self.normalise_abs = np.abs(self.normalise_field)  # Store these now for speed
        self.normalisePhase = np.angle(self.normalise_field)

    ##################### PHASE #############################################

    def set_stable_ROI(self, roi):
        """Set the location of the the ROI used for maintaining a constant
        background phase, i.e. this should be a background region of the image.
        The roi should be an instance of the Roi class.
        """
        assert isinstance(roi, Roi), "Argument must be an instance of Roi."
        self.stable_roi = roi

    def set_relative_phase(self, relative_phase):
        """Sets whether or not to use relative phase, i.e. phase
        is relative to the phase of the background hologram.
        """

        assert relative_phase == True or relative_phase == False, (
            "Argument of set_relative_phase must be True or False"
        )
        self.relative_phase = relative_phase

    ##################### REFOCUSING #########################################

    def set_depth(self, depth):
        """Set the depth for numerical refocusing.
        Arguments:
            depth       : int or float
                        Depth to refocus at in the same units as pixel_size and wavelength.
        """
        assert isinstance(depth, (int, float)), (
            "Argument of set_depth must be an int or float."
        )
        self.depth = depth

    def update_propagator(self, img):
        """Create or re-create the propagator using current parameters. img
        should be an 2D numpy array of the same size as the images to be processed.
        """

        if self.mode == self.INLINE_MODE:
            assert self.pixel_size is not None, (
                "Pixel size must be specified before propagator is created."
            )
            self.propagator_pixel_size = self.pixel_size * self.downsample
            downsample = self.downsample
        else:
            assert self.oa_pixel_size is not None, (
                "Pixel size must be specified before propagator is created."
            )
            self.propagator_pixel_size = self.oa_pixel_size
            downsample = 1  # The way oa_pixel_size is calculated, we already take account of the downsample factor

        prop_width = int(np.shape(img)[1] / downsample / 2) * 2
        prop_height = int(np.shape(img)[0] / downsample / 2) * 2

        self.propagator = propagator(
            (prop_height, prop_width),
            self.wavelength,
            self.propagator_pixel_size,
            self.depth,
            precision=self.precision,
            geometry=self.geometry,
            use_numba=self.use_numba,
        )

        # If using CUDA we send propagator to GPU now to speed up refocusing later
        if self.cuda and cuda_available:
            self.propagator.propagator = cp.array(self.propagator.propagator)

    def set_downsample(self, downsample):
        """Set the downsample factor. This will cause the propagator to be
        recreated when next needed, call update_propagator to force this immediately.
        Arguments:
            downsample   : int or float
                           Factor to downsample the image by, must be > 0.
                           If > 1, the image will be downsampled by this factor.
        """
        if downsample != self.downsample:
            self.propagator = None  # Force to be recreated when needed

        self.downsample = downsample

    ############### AUTO FOCUS ###################################################

    def set_find_focus_parameters(self, **kwargs):
        """Sets the parameters used by the find_focus method.

        Keyword Arguments:
            depth_range   : double
                           tuple of (min, max) depths to search within in m.
            roi          : instance of Roi
                           area to assess focus within, default is None in which
                           case all of image is used.
            method       : str
                           focus metric to use.
            margin       : int
                           if specified only the Roi and a margin will be
                           refocused. If None (default) the whole image will be
                           refocused regardless. Has no effect if roi not specified.
            coarse_search_interval  : Number of points to check explicitly before
                                    optimising. Default is None, in which case
                                    this is not performed.
            use_prop_lut : bool
                                Whether or not to use the auto focus propagator LUT for 
                                finding focus. Default is False. If True, the LUT
                                must have been generated with make_auto_focus_propagator_LUT() 
                                and must cover the depth range specified by depth_range.
        """
        self.find_focus_depth_range = kwargs.get("depth_range", (0, 0.1))
        self.find_focus_roi = kwargs.get("roi", None)
        self.find_focus_method = kwargs.get("method", "sum")
        self.find_focus_margin = kwargs.get("margin", None)
        self.coarse_search_interval = kwargs.get("coarse_search_interval", None)
        self.use_auto_focus_propagator_lut = kwargs.get("use_prop_lut", False)

    def find_focus(self, img):
        """Automatically finds the best focus position for hologram 'img'
        using parameters previously defined (such as wavelength) as well
        as parameters set by set_find_focus_parameters().

        Arguments:
            img         : ndarray
                          2D array containing hologram

        Returns:
            float       : optimal refocus depth
        """

        if self.use_auto_focus_propagator_lut:
            if self.auto_focus_propagator_lut is not None:
                if (
                    self.auto_focus_propagator_lut.wavelength != self.wavelength
                    or self.auto_focus_propagator_lut.pixel_size != self.pixel_size
                ):
                    raise "Auto focus propagator LUT was generated with different wavelength or pixel size to current settings."
            else:
                raise "Auto focus propagator LUT has not been generated, but use_propagator_LUT is set to True."
            prop_lut = self.auto_focus_propagator_lut
        else:
            prop_lut = None

        args = {
            "background": self.background,
            "window": self.window,
            "roi": self.find_focus_roi,
            "margin": self.find_focus_margin,
            "numba": numba_available and self.use_numba,
            "cuda": cuda_available and self.cuda,
            "prop_lut": prop_lut,
            "coarse_search_interval": self.find_focus_coarse_search_interval,
            "correct_curvature": self.correct_curvature,
            "source_distance": self.source_distance,
        }

        return find_focus(
            img,
            self.wavelength,
            self.pixel_size,
            self.find_focus_depth_range,
            self.find_focus_method,
            precision=self.precision,
            **args,
        )

    def auto_focus(self, img):
        """Finds the best focus for hologram 'img' using parameters previously
        defined (such as wavelength) as well as by set_find_focus_parameters(),
        and then refocuses to this depth and returns the refocused image. Note that
        this is sets the depth of the Holo class to the found focus depth, so 
        subsequent calls to process() will be at this depth unless depth is 
        changed again.
        
        Arguments:
            img         : ndarray
                          2D array containing hologram
        Returns:
            ndarray      : refocused image at optimal focus depth.
        """
        focus_depth = self.find_focus(img)
        self.set_depth(focus_depth)
        return self.__process_inline(img)    

    

    def make_propagator_LUT(self, dimension, depth_range, num_depths):
        """Creates and stores a LUT of propagators for faster refocus.

        Arguments:
            dimension   : int or (int, int) or ndarray
                          dimension of hologram to determine size of propagators in LUT.
            depth_range : tuple
                          depths range to create propagators for: (min depth, max depth)
            num_depths  : int
                          number of depths to create propagators for, evenly spaced within depth_range/
        """
        self.propagator_lut = PropLUT(
            dimension,
            self.wavelength,
            self.pixel_size,
            depth_range,
            num_depths,
            use_numba=(numba_available and self.use_numba),
            precision=self.precision,
        )

    def make_auto_focus_propagator_LUT(self, dimension, depth_range, num_depths, roi = None, margin = None):
        """Creates and stores a LUT of propagators for faster finding of focus.

        Arguments:
            dimension   : int or (int, int) or ndarray
                          dimension of hologram to determine size of propagators in LUT.
            depth_range : tuple
                          depths range to create propagators for: (min depth, max depth)
            num_depths  : int
                          number of depths to create propagators for, evenly spaced within depth_range.
            roi         : instance of Roi
                          area that focus will be assessed within, default is None. This only
                          matters for the propagator if a margin is specified, in which
                          case only the roi and a margin around it will be refocused to assess focus, 
                          so the propagators only need to be the size of this.
            margin      : int
                          if specified only the roi and a margin will be refocused to assess focus, 
                          default is None in which case the whole image will be refocused to assess focus. 
                          If margin is specified, the propagators in the LUT will only be the size of the 
                          roi plus the margin, to speed up refocusing for finding focus.
        """
        dimensions = propagator_size_for_roi(dimension, margin = margin, roi = roi)
        self.auto_focus_propagator_lut = PropLUT(
            dimensions,
            self.wavelength,
            self.pixel_size,
            depth_range,
            num_depths,
            use_numba=(numba_available and self.use_numba),
            precision=self.precision,
        )

    def clear_propagator_LUT(self):
        """Deletes the LUT of propagators."""
        self.propagator_lut = None

    def clear_auto_focus_propagator_LUT(self):
        """Deletes the LUT of propagators."""
        self.auto_focus_propagator_lut = None

    ################### DEPTH STACK ##############################################

    def depth_stack(self, img, depth_range, num_depth):
        """Create a depth stack using current parameters, producing a set of
        'num_depth' refocused images equally spaced within depth_range.

        Arguments:
            img        : ndarray
                         hologram
            depth_range : tuple
                         depths to focus to: (min depth, max depth)
            num_depth    : int
                         number of depths to create images for within depth_range

        Returns:
            FocusStack : instance of FocusStack containing the refocused images.
        """

        if self.mode == self.INLINE_MODE:
            pre_background = self.background
            post_background = None
        else:
            pre_background = None
            post_background = None
        args = {
            "background": self.background,
            "window": self.window,
            "numba": numba_available and self.use_numba,
        }

        return refocus_stack(
            img,
            self.wavelength,
            self.pixel_size,
            depth_range,
            num_depth,
            precision=self.precision,
            **args,
        )

    ########################### GENERAL SETTINGS ################################

    def set_use_cuda(self, use_cuda):
        """Set whether to use GPU if available, use_cuda is True to use GPU or
        False to not use GPU.
        """
        self.cuda = use_cuda

    def set_use_numba(self, use_numba):
        """Set whether to use Numba JIT if available, use_numba is True to use
        Numba or False to not use Numba.
        """
        self.use_numba = use_numba

    def set_precision(self, precision):
        """Sets whether to use single or double precision."""
        assert precision == "single" or precision == "double", (
            "Precision must be 'single' or 'double'."
        )
        self.precision = precision
        if self.precision == "double":
            self.imType = "float64"
        else:
            self.imType = "float32"

    def __str__(self):
        return (
            "PyHoloscope Holo Class. Wavelength: "
            + str(self.wavelength)
            + ", Pixel Size: "
            + str(self.pixel_size)
        )



"""
PyHoloscope - Fast Holographic Microscopy for Python

The Propagator class is a container to hold a propagator for wavefront
propagation in holography along with metadata. It does not generate the
propagator.
"""

import numpy as np


class Propagator:
    """Represents a propagator for wavefront propagation in holography.

    Attributes:

        propagator  : 2D complex numpy array containing the propagator.

        wavelength  : The wavelength of the light used in the propagation (m).

        pixel_size  : The size of the pixels in the hologram (m).

        depth       : Propagation distance (m).
    """

    wavelength = None
    pixel_size = None
    depth = None
    propagator = None
    shape = None
    geometry = None

    def __init__(
        self, propagator, wavelength=None, pixel_size=None, depth=None, geometry=None
    ):
        """
        Initializes the Propagator instance by stoing the propagator and its attributes.

        Parameters:
            propagator: 2D complex numpy array containing the propagator.

        Optional Keyword Arguments:
            wavelength: float
                        The wavelength of the light (m).
            pixel_size: float
                        The size of the pixels in the hologram (m).
            depth: float
                   Propagation distance (m).
            geometry: str
                      Geometry of the propagator ('plane' or 'point').
        """
        self.propagator = propagator
        self.wavelength = wavelength
        self.pixel_size = pixel_size
        self.depth = depth
        self.geometry = geometry
        self.shape = np.shape(propagator)

    def has_attributes(
        self, wavelength=None, pixel_size=None, depth=None, geometry=None
    ):
        """Checks if the propagator has the specified attributes.

        Keyword arguments:
            wavelength      : float
                              The wavelength to check against (m).
            pixel_size      : float
                              The pixel size to check against (m).
            depth           : float
                              The depth to check against (m).
            geometry        : str
                              The geometry to check against ('plane' or 'point').
        Returns:
            bool            : True if the propagator has all the specified attributes,
                              False otherwise.
        """

        if wavelength is not None and self.wavelength != wavelength:
            return False
        if pixel_size is not None and self.pixel_size != pixel_size:
            return False
        if depth is not None and self.depth != depth:
            return False
        if geometry is not None and self.geometry != geometry:
            return False

        return True



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

This file contains functions for working with off-axis holograms.
"""

import math
import numpy as np
import scipy

import matplotlib.pyplot as plt

try:
    import cupy as cp

    cuda_available = True
except:
    cuda_available = False

from pyholoscope.utils import extract_central, dimensions


def off_axis_demod_pixel_size(hologram, pixel_size, crop_radius):
    """Return effective pixel size after off-axis demodulation.

    The off-axis demodulated field is smaller than the hologram by a factor
    determined by the crop radius. Where a non-square crop area is used,
    the vertical dimensions are used - in practice this should not matter
    as the aspect ratio of the crop area and the full hologram are normally
    the same.

    Arguments:
        hologram       : tuple or ndarray
                        shape of the raw hologram (or the hologram itself)
        pixel_size     : float
                        pixel size of the raw hologram
        crop_radius    : int or (int, int)
                        crop radius used for off-axis demodulation
     
    Returns:
        float : pixel size in the demodulated field
    """

    h, _ = dimensions(hologram)
    crop_radius = dimensions(crop_radius)[0]
    if crop_radius == 0:
        raise ValueError("crop_radius must be non-zero")
    demod_height = int(crop_radius) * 2
    return pixel_size / float(demod_height) * float(h)



def off_axis_demod(
    hologram,
    crop_centre,
    crop_radius,
    real_fft=False,
    return_full=False,
    return_fft=False,
    mask=None,
    cuda=False,
):
    """Removes spatial modulation from off-axis hologram to obtain complex field.

    By default, returns the complex field as a 2D numpy array of size
    2 * crop_radius. If return_full is True, the returned
    array will instead by the same size as the input hologram. If return_fft is
    True, function returns a tuple (field, FFT) where FFT is a log scaled
    image of the FFT of the hologram (2D numpy array, real).

    Arguments:
          hologram   : numpy.ndarray
                       2D numpy array, real, raw hologram
          crop_centre : tuple of (int, int).
                       pixel location in FFT of modulation frequency (y, x)
          crop_radius : int or (int, int)
                       semi-diameter of sqaure or rectangle to extract
                       around modulation frequency. Provide a single int
                       for a sqaure area or a tuple of (h,w) for a rectangle.

    Keyword Arguments:
          real_fft    : boolean
                        if True, the real FFT will be used for speed up (default is False).
                        This only works
                        if the reference is tilted so that the modulation is at an angle of
                        approximately 45 degrees to the x and y axes, so that the shifted
                        (modulated) signal is in one quadrant of the FFT.
          return_full : boolean
                       if True, the returned reconstruction will be the same
                       size as the input hologram, otherwise it will be
                       2 * crop_radius. (Default is False)
          return_fft  : boolean
                       if True will return a tuple of (demod image,
                       log scaled FFT) for display purposes
                       (Default is False)
          mask       : ndarray
                       2D complex array. Custom mask to use around
                       demodulation frequency. Must match size of
                       (crop_radius_x, crop_radius_y)
          cuda      :  boolean
                       if True GPU will be used if available.
    Returns:
          numpy. ndarray   : reconstructed field as complex numpy array or
                             tuple of (ndarray, ndarray) if return_fft is True
    """

    # Size of image in pixels
    height, width = np.shape(hologram)

    crop_centre = dimensions(crop_centre)
    crop_radius = dimensions(crop_radius)

    # Apply 2D FFT
    if cuda is False or cuda_available is False:
        if real_fft:
            camera_fft = scipy.fft.rfft2(hologram)
        else:
            camera_fft = scipy.fft.fftshift(scipy.fft.fft2(hologram))
    else:
        if type(hologram) is np.ndarray:
            hologram = cp.array(hologram)
        if type(mask) is np.ndarray:
            mask = cp.array(mask)
        if real_fft:
            camera_fft = cp.fft.rfft2(cp.array(hologram))
        else:
            camera_fft = cp.fft.fftshift(cp.fft.fft2(cp.array(hologram)))

    # Shift the ROI to the centre
    shifted_fft = camera_fft[
        round(crop_centre[0] - crop_radius[0]) : round(crop_centre[0] + crop_radius[0]),
        round(crop_centre[1] - crop_radius[1]) : round(crop_centre[1] + crop_radius[1]),
    ]

    # Apply the mask
    if mask is not None:
        assert np.shape(mask) == np.shape(shifted_fft), (
            f"Incorrect mask size, mask is {np.shape(mask)}, fft region is {np.shape(shifted_fft)}."
        )
        masked_fft = shifted_fft * mask
    else:
        masked_fft = shifted_fft

    if return_full:
        h, w = np.shape(hologram)
        (
            h2,
            w2,
        ) = np.shape(masked_fft)
        x, y = round((w - w2) / 2), round((h - h2) / 2)
        output = np.zeros((h, w), dtype="complex")
        output[y : y + h2, x : x + w2] = masked_fft
        masked_fft = output

    # Reconstruct complex field
    if cuda is False or cuda_available is False:
        recon_field = scipy.fft.ifft2(scipy.fft.fftshift(masked_fft))
    else:
        recon_field = cp.asnumpy(cp.fft.ifft2(cp.fft.fftshift(masked_fft)))

    if return_fft:
        if cuda is True and cuda_available is True:
            try:
                camera_fft = cp.asnumpy(camera_fft)
            except:
                pass
        return recon_field, np.log(np.abs(camera_fft) + 0.000001)  # Stops log(0)

    else:
        return recon_field


def off_axis_find_mod(hologram, mask_fraction=0.1, real_fft=False):
    """Finds the location of the off-axis holography modulation peak in the FFT.

    Arguments:
          hologram     : ndarray
                         2D numpy array, real, raw hologram

    Keyword Arguments:
          mask_fraction : float
                         between 0 and 1, fraction of image around d.c. to
                         mask to avoid the d.c. peak being detected
                         (default = 0.1).
          real_fft    : bool
                          if True, then the real FFT will be used (default is False). This only
                          works if the reference is tilted so that the modulation is at
                          an angle of appoximately 45 degrees to the x and y axes, so that
                          the shifted (modulated) signal is in one quadrant of the FFT.
    Returns:
          tuple of (int, int), modulation location in FFT (x location, y location)
    """

    # Apply 2D FFT
    if real_fft:
        camera_fft = np.abs(scipy.fft.rfft2((hologram)))

        # Need to crop out DC otherwise we will find that. Set the areas around
        # dc (for both quadrants) to zero. The size of the masked area is mask_fraction * the
        # size of the image (smallest dimension)
        maskSize = int(np.min(np.shape(hologram)) * mask_fraction)

        camera_fft[:maskSize, :maskSize] = 0
        camera_fft[-maskSize:, :maskSize] = 0

    else:
        camera_fft = np.abs(scipy.fft.fftshift(scipy.fft.fft2(hologram)))

        # Need to crop out DC otherwise we will find that. Set the areas around
        # dc to zero. The size of the masked area is
        # mask_fraction * the size of the image (smallest dimension)
        maskSize = int(np.min(np.shape(hologram)) * mask_fraction)
        cy, cx = np.shape(camera_fft)[:2]
        camera_fft[
            cy // 2 - maskSize : cy // 2 + maskSize,
            cx // 2 - maskSize : cx // 2 + maskSize,
        ] = 0
        camera_fft[:, : cx // 2] = 0

    peak_location = np.unravel_index(camera_fft.argmax(), camera_fft.shape)

    return peak_location[0], peak_location[1]


def off_axis_find_crop_radius(hologram, mask_fraction=0.1, real_fft=False):
    """Estimates the off-axis crop radius based on modulation peak position. If the
    hologram is square, this is the radius of a circle, otherwise if it is rectangular
    than the crop radius is a tuple of (y radius, x radius), corresponding to
    half the lengths of the two axes of an ellipse.

    Arguments:
          hologram     : numpy.ndarray
                         raw hologram, 2D real array

    Keyword Arguments:
          mask_fraction : float
                         between 0 and 1, fraction of image around d.c. to
                         mask to avoid the d.c. peak being detected
                         (default = 0.1).
            real_fft    : bool
                          if True, then the real FFT will be used (default is False). This only
                          works if the reference is tilted so that the modulation is at
                          an angle of appoximately 45 degrees to the x and y axes, so that
                          the shifted (modulated) signal is in one quadrant of the FFT.
    Returns:
          tuple of (int, int) = (x radius, y radius)
    """

    h = np.shape(hologram)[0]
    w = np.shape(hologram)[1]

    peak_y, peak_x = off_axis_find_mod(
        hologram, mask_fraction=mask_fraction, real_fft=real_fft
    )

    # The crop radii will have the same ratio as the width and height of the hologram
    aspect_ratio = h / w

    if real_fft:
        peak_y_adj = min(peak_y, np.abs(h - peak_y))
        peak_x_adj = min(peak_x, np.abs(w - peak_x))
    else:
        peak_y_adj = np.abs(h // 2 - peak_y)
        peak_x_adj = np.abs(w // 2 - peak_x)

    peak_y_adj = peak_y_adj / aspect_ratio

    mod_freq = np.sqrt(peak_x_adj**2 + peak_y_adj**2)
    crop_radius = mod_freq / 3

    # If the crop radius will result in a ROI larger than the image, adjust it
    crop_radius = min(
        crop_radius,
        (h - peak_y) / aspect_ratio,  # check bottom
        peak_x,  # left
        w - peak_x,  # right
        peak_y / aspect_ratio,  # top
    )

    crop_radius_x = int(round(crop_radius))
    crop_radius_y = int(round(crop_radius * aspect_ratio))

    return crop_radius_y, crop_radius_x


def off_axis_predict_mod(
    wavelength, pixel_size, num_pixels, tilt_angle, rotation=0, real_fft=False
):
    """Predicts the location of the modulation peak in the FFT.

    Arguments:
          wavelegnth   : float
                         light wavelength in metres
          pixel_size    : float
                         hologram physical pixel size in metres
          num_pixels    : int or (int, int)
                         hologram size in pixels. If rectangular, provide a tuple of (h,w) or a 2D numpy array
                         the same shape as the hologram.
          tilt_angle    : float
                         angle of reference beam on camera in radians

    Keyword Arguments:
          rotation     : float
                         rotation of tilt with respect to x axis, in radians (default is 0)
            real_fft    : bool
                         if True, then the position will be correct for a real FFT.

    Returns:
          tuple of (int, int), location of modulation (x pixel, y pixel)

    """

    # Spatial frequency of modulation
    ref_freq = math.sin(tilt_angle) / wavelength

    # Spatial frequency at edge of FFT
    max_spatial_freq = 1 / (pixel_size * 2)

    im_size_y, im_size_x = dimensions(num_pixels)

    # Pixel corresponding to frequency in Fourier Domain
    if rotation % math.pi < math.pi / 2:
        mod_freq_px_x = round(
            ref_freq / max_spatial_freq * np.abs(np.cos(rotation)) * im_size_x / 2
        )
        mod_freq_px_y = round(
            ref_freq / max_spatial_freq * np.abs(np.sin(rotation)) * im_size_y / 2
        )
    else:
        mod_freq_px_x = round(
            ref_freq
            / max_spatial_freq
            * np.abs(np.cos(math.pi - rotation))
            * im_size_x
            / 2
        )
        mod_freq_px_y = im_size_y - round(
            ref_freq / max_spatial_freq * np.abs(np.sin(rotation)) * im_size_y / 2
        )

    if mod_freq_px_x < 0:
        mod_freq_px_x = mod_freq_px_x + im_size_x
    if mod_freq_px_y < 0:
        mod_freq_px_y = mod_freq_px_y + im_size_y

    # We want the position in the fftshifted FFT, so we need to adjust the coordinates
    mod_freq_px_x = (mod_freq_px_x + im_size_x // 2) % im_size_x
    mod_freq_px_y = (mod_freq_px_y + im_size_y // 2) % im_size_y

    return mod_freq_px_y, mod_freq_px_x


def off_axis_predict_mod_distance(
    wavelength, pixel_size, num_pixels, tilt_angle, rotation=0
):
    """Predicts the absolute distance of the modulation peak in the FFT from the dc.

    Arguments:
          wavelegnth   : float
                         light wavelength in metres
          pixel_size    : float
                         hologram physical pixel size in metres
          num_pixels    : int or (int, int)
                         hologram size in pixels,
          tilt_angle    : float
                         angle of reference beam on camera in radians

    Keyword Arguments:
          rotation     : float
                         rotation of tilt with respect to x axis, in radians (default is 0)

    Returns:
          float        : distance in pixels

    """

    x, y = off_axis_predict_mod(
        wavelength, pixel_size, num_pixels, tilt_angle, rotation
    )

    return math.sqrt(x**2 + y**2)


def off_axis_predict_tilt_angle(hologram, wavelength, pixel_size, mask_fraction=0.1):
    """Returns the reference beam tilt based on the hologram modulation. The angle
    is returned in radians.

    Arguments:
          hologram     : ndarray
                         2D numpy array, real, hologram
          wavelength   : float
                         light wavelength in metres
          pixel_size    : float
                         hologram physical pixel size in metres

    Optional Keyword Arguments:
          mask_fraction : float
                         between 0 and 1, fraction of image around d.c. to
                         mask to avoid the d.c. peak being detected
                         (default = 0.1).

    Returns:
          float        : tilt angle in radians
    """

    # Wavenumber
    k = 2 * math.pi / wavelength

    h, w = np.shape(hologram)[:2]

    # Find the location of the peak
    peak_location = off_axis_find_mod(
        hologram, mask_fraction=mask_fraction, real_fft=True
    )

    # Pixel sizes in FFT (the spatial frequency)
    v_pixel_spatial_freq = 1 / (pixel_size * np.shape(hologram)[0])
    h_pixel_spatial_freq = 1 / (pixel_size * np.shape(hologram)[1])

    # Depending on quadrant could be relative to either top-left or
    # top-right corner, so check both and use the closest distance
    peak_dist1 = np.sqrt(
        (v_pixel_spatial_freq * peak_location[0]) ** 2
        + (h_pixel_spatial_freq * peak_location[1]) ** 2
    )
    peak_dist2 = np.sqrt(
        (v_pixel_spatial_freq * peak_location[0]) ** 2
        + (h_pixel_spatial_freq * (peak_location[1] - w)) ** 2
    )
    spatial_freq = min(peak_dist1, peak_dist2)

    tilt_angle = math.asin(2 * math.pi * spatial_freq / k)

    return tilt_angle



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

This file contains functions for handling and displaying phase
maps.

"""

import warnings
import time
import math

import numpy as np
import scipy.optimize

from skimage.restoration import unwrap_phase

from PIL import Image

import cv2 as cv

from pyholoscope.roi import Roi
from pyholoscope.focus_stack import FocusStack
from pyholoscope.prop_lut import PropLUT
from pyholoscope.utils import extract_central, phase


def relative_phase(img, background):
    """Removes global phase from image using reference (background) field.

    The function works on both fields (complex arrays) and phase maps (real
    arrays), and returns a corrected field/phase map of the same type.

    Parameters:
          img        :  numpy.ndarray
                        2D numpy array, real or complex. If real
                        it is taken to be phase map, otherwise if complex
                        it is the field.
          background :  numpy.ndarray
                        2D numpy array, real or complex. Backround phase/field
                        to subtract. Must be same type and size as img.

    Returns:
          numpy.ndarray : float or complex
    """

    assert np.iscomplexobj(img) == np.iscomplexobj(background), (
        "img and background must both be real or both be complex"
    )
    assert np.shape(img) == np.shape(background), (
        "img and background must be the same size."
    )

    # If both inputs are phases, simply subtract
    if not np.iscomplexobj(img) and not np.iscomplexobj(background):
        return img - background

    # If both input are fields, divide
    if np.iscomplexobj(img) and np.iscomplexobj(background):
        return img / background * np.abs(background)


def relative_phase_self(img, roi=None):
    """Makes the phase in an image relative to the mean phase in either the
    whole image or a specified ROI of the image.

    The function works on either fields (complex arrays) and phase maps (float
    arrays), and returns a corrected field/phase map of the same type.

    Parameters:
          img        :  numpy.ndarray
                        2D array, real or complex. If real
                        it is taken to be phase map, otherwise if complex
                        it taken to be the complex field.

    Keyword Arguments:
          roi        :  pyholoscope.Roi
                        region of interest to make phase relative to. In the output image
                        the mean phase in this region will be zero.
                        (default = None)

    Returns:
          numpy.ndarray : float or complex
    """

    if roi is None:
        av_phase = mean_phase(img)
    else:
        av_phase = mean_phase(roi.crop(img))

    if np.iscomplexobj(img):
        out_image = img * np.exp(1j * -1 * av_phase)
    else:
        out_image = img - av_phase

    return out_image


def obtain_tilt(img):
    """Estimates the global tilt in the 2D unwrapped phase.

    This can be used to correcte tilts in the phase due to, for example, a
    tilt in the coverglass. Returns a phase map approximating the tilt as
    a 2D real numpy array.

    Parameters:
          img        :  ndarray
                        2D numpy array, real. Unwrapped phase.

    Returns:
          numpy.ndarray : 2D real array, map of tilt.

    """

    # If there is a tilt then there will be a phase gradient across the image
    tilt_x, tilt_y = np.gradient(img)
    tilt_x = np.mean(tilt_x)
    tilt_y = np.mean(tilt_y)

    mx, my = np.indices(np.shape(img))

    tilt = mx * tilt_x + my * tilt_y

    residual = np.median(img - tilt)

    tilt = tilt + residual

    return tilt


def phase_unwrap(img):
    """Returns unwrapped version of 2D phase map.

    Parameters:
          img        :  ndarray
                        2D numpy array. Either complex (the complex field) or
                        float (the wrapped phase).
    Returns:
          numpy.ndarray : 2D array of floats, unwrapped phase.

    """

    if np.iscomplexobj(img):
        img = phase(img)

    img = unwrap_phase(img)

    return img


def synthetic_DIC(img, shear_angle=math.pi):
    """Generates a simple, non-rigorous DIC-style image for display.

    The ouput should appear similar to a relief map, with dark and light
    regions correspnding to positive and negative phase gradients along the
    shear angle direction (default is horizontal = 0 rad). The input must
    be the complex field, not a phase map.

    Parameters:
          img        :  ndarray
                        2D numpy array, complex, the field.

    Keyword Arguments:
          shear_angle :  float
                        angle in radians of the shear direction.
                        (default = pi)

    Returns:
          numpy.ndarray : 2D arrays of floats, the DIC image.
    """

    # Calculate gradient on original image and image phase shifted by pi. Using
    # the smallest phase gradient avoids effects due to phase wrapping
    if not np.iscomplexobj(img):
        img = np.exp(1j * img)  # Convert phase map to field
    sobelC1 = phase_gradient_dir(img)
    sobelC2 = phase_gradient_dir(img * np.exp(1j * math.pi))

    use1 = np.abs(sobelC1) < np.abs(sobelC2)

    sobelC1[np.invert(use1)] = 0
    sobelC2[use1] = 0
    sobelC = sobelC1 + sobelC2

    # Rotate the gradient to shear angle
    sobelC = sobelC * np.exp(1j * shear_angle)

    # DIC is product of phase gradient along one direction and image intensity
    DIC = np.real(sobelC) * (np.max(np.abs(img)) - np.abs(img))

    return DIC


def phase_gradient_dir(img):
    """Returns the directional phase gradient of an image.

    The output is a complex 2D numpy array, with the horizontal and vertical
    gradients encoded in the real and imaginary parts.

    Parameters:
          img        :  ndarray
                        2D numpy array, real or complex, the field or phase map
    Returns:
          numpy.ndarray : 2D complex array, phase gradient image.
    """

    if np.iscomplexobj(img):
        img = np.angle(img)

    # Phase gradient in x and y directions
    sobelx = cv.Sobel(img, cv.CV_64F, 1, 0)
    sobely = cv.Sobel(img, cv.CV_64F, 0, 1)

    # Encode in complex array
    sobelC = sobelx + 1j * sobely

    return sobelC


def phase_gradient_amp(img):
    """Returns the amplitude of the phase gradient of an image.

    This isn't very useful by itself if applied to wrapped phase as it
    find an edge whenever the phase is wrapped, phase_gradient avoids this
    problem.

    Parameters:
          img        :  ndarray
                        2D numpy array, real or complex, the field or phase map
    Returns:
          numpy.ndarray : 2D arrays of floats, maps of amplitudes of phase gradient.
    """

    # If we are given the field, calculate the phase map
    if np.iscomplexobj(img):
        img = np.angle(img)

    # Phase gradient in x and y directions
    sobelx = cv.Sobel(img, cv.CV_64F, 1, 0)  # Find x and y gradients
    sobely = cv.Sobel(img, cv.CV_64F, 0, 1)

    # Return amplitude
    return np.sqrt(sobelx**2 + sobely**2)


def phase_gradient(img):
    """Returns the amplitude of the phase gradient of an image.

    This function is able to handle wrapped phase without finding an edge
    at the wrap points. Returns a 2D numpy array, real containing the amplitude
    of the gradient at each point.

    Parameters:
          img        :  ndarray
                        2D numpy array, real or complex, the field or phase map

    Returns:
          numpy.ndarray : 2D array of floats, maps of amplitudes of phase gradient.
    """

    # We calculate the phase gradient twice, once adding a pi phase shift. If
    # the phase has wrapped at a point, this will produce a much bigger phase
    # gradient which we remove by taking the minimum of the two gradients.

    phase_grad1 = phase_gradient_amp(img)

    if np.iscomplexobj(img):
        phase_grad2 = phase_gradient_amp(img * np.exp(1j * math.pi))
    else:
        phase_grad2 = phase_gradient_amp(np.exp(1j * img))

    return np.minimum(phase_grad1, phase_grad2)


def mean_phase(img):
    """Returns the mean phase of a field or phase map.

    Parameters:
          img        :  ndarray
                        2D array, real or complex, the field or phase map

    Returns:
          float      : mean phase of the image.
    """

    if np.iscomplexobj(img):
        return np.angle(np.sum(img))  # The way to average the phase in a field
    else:
        return np.mean(img)



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

Roi: Class for region of interest.
"""


class Roi:
    def __init__(self, x, y, width, height):
        """Initialise ROI.

        Arguments:
            x, y    : int
                      x and y coordinates of top-left corner
            width   : int
                      width of ROI
            height  : int
                      height of ROI
        """
        self.x = max(int(x), 0)
        self.y = max(int(y), 0)
        self.width = max(int(width), 0)
        self.height = max(int(height), 0)

    def __str__(self):
        return (
            str(self.x)
            + ","
            + str(self.y)
            + ","
            + str(self.width)
            + ","
            + str(self.height)
        )

    def constrain(self, min_x, min_y, max_x, max_y):
        """Stops ROI exceeding a specified size by adjusting ROI coordinates and size.

        Arguments:
            min_x    : int
                       minimum x coordinate
            min_y    : int
                       minimum y coordinate
            max_x    : int
                       maximum x coordinate
            max_y    : int
                       maximum y coordinate
        """

        self.x = max(self.x, min_x, 0)
        self.y = max(self.y, min_y, 0)

        self.width = max(min(self.width, max_x - self.x), 0)
        self.height = max(min(self.height, max_y - self.y), 0)

    def crop(self, img):
        """Crop and image using the ROI.

        Arguments:
            img : numpy.ndarray
                  input image

        Returns:
            numpy.ndarray : cropped image
        """
        return img[self.y : self.y + self.height, self.x : self.x + self.width]

    def clear_outside(self, img):
        """Set pixels in img to be zero if outside ROI.

        Arguments:
            img : numpy.ndarray
                  input image

        Returns:
            numpy.ndarray : image with pixels outside ROI set to zero

        """
        imgOut = img.copy()
        imgOut[: self.y, :] = 0  # set pixels above ROI to zero
        imgOut[self.y + self.height :, :] = 0  # set pixels below ROI to zero
        imgOut[:, : self.x] = 0  # set pixels to the left of ROI to zero
        imgOut[:, self.x + self.width :] = 0  # set pixels to the right of ROI to zero

        return imgOut

    def clear_inside(self, img):
        """Set pixels in img to be zero if inside ROI.

        Arguments:
            img : numpy.ndarray
                  input image

        Returns:
            numpy.ndarray : image with pixels inside ROI set to zero
        """

        img[self.y : self.y + self.height, self.x : self.x + self.width] = 0

        return img



# -*- coding: utf-8 -*-
"""
PyHoloscope - Python package for holgoraphic microscopy

The Sim package provide utilites for simulating off-axis holograms.

"""

import math

import numpy as np

from pyholoscope import dimensions


def off_axis(
    object_field, wavelength, pixel_size, tilt_angle, rotation=math.pi / 4, OPD=0
):
    """Generates simulated off-axis hologram.

    Arguments:
        object_field : ndarray, complex
                      complex field representing object
        wavelength  : float
                      wavelength of simulated light source
        pixel_size   : float
                      real size of pixels in hologram
        tilt_angle   : float
                      angle of reference beam on camera in radians

    Keyword Arguments:
        rotation    : float
                      rotation of the tilt with respect to x axis in rad
                      (Default is pi/4 = 45 deg)
        OPD         : optical path difference between beams (Default is 0)

    """

    # Convert wavelength to wavenumber
    k = 2 * math.pi / wavelength

    # Check size of object
    num_points_y, num_points_x = np.shape(object_field)

    # Generate tilted reference field
    ref_field = np.ones((num_points_y, num_points_x)) + 1j * np.ones(
        (num_points_y, num_points_x)
    )
    (xM, yM) = np.meshgrid(range(num_points_x), range(num_points_y))

    # Effect of tilt
    ref_freq = k * math.sin(tilt_angle)
    ref_field = ref_field * np.exp(
        1j * (ref_freq * pixel_size * (xM * np.cos(rotation) + yM * np.sin(rotation)))
    )

    # Effect of OPD
    ref_field = ref_field * np.exp(1j * 2 * math.pi * OPD / wavelength)

    # Field at camera is superposition of images of object and reference.
    # We are assuming 1:1 magnification here.
    cam_field = ref_field + object_field

    # Intensity at camera is square of field,
    cam_intensity = np.abs(cam_field) ** 2

    return cam_intensity



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

PropLUT is the class for generating and storing propagator look up tables.

"""

import numpy as np

from pyholoscope.focusing import propagator
from pyholoscope.focusing_numba import propagator_numba
from pyholoscope.utils import dimensions
from pyholoscope.propagator import Propagator


class PropLUT:
    """Stores a propagator look up table (LUT).

    The LUT contains angular spectrum propagators for the specified parameters.

    """

    def __init__(
        self,
        img_size,
        wavelength,
        pixel_size,
        depth_range,
        num_depths,
        geometry="plane",
        use_numba=True,
        precision="single",
    ):
        """Creates a propagator look up table (LUT) containing angular spectrum propagators.

        Arguments:
            img_size   : int or tuple of (int, int)
                        size of propagators, square or rectangular
            wavelength: float
                        wavelength of light
            pixel_size : float
                        physical size of pixels
            depth_range: tuple
                        range of depths to generate propagators for
            num_depths : int
                        number of depths to generate propagators for

        Keyword Arguments:
            geometry  : str
                        'plane' (default) or 'point'
            numba     : bool
                        flag to use numba for speed up (default = False)
            precision : str
                        'single' or 'double' (default = 'single')
        """

        if precision == "double":
            dataType = "complex128"
        else:
            dataType = "complex64"

        self.depths = np.linspace(depth_range[0], depth_range[1], num_depths)
        self.size = img_size
        self.num_depths = num_depths
        self.wavelength = wavelength
        self.pixel_size = pixel_size
        self.geometry = geometry
        h, w = dimensions(img_size)
        self.prop_table = np.zeros((num_depths, h, w), dtype=dataType)
        for idx, depth in enumerate(self.depths):
            self.prop_table[idx, :, :] = propagator(
                (h, w),
                wavelength,
                pixel_size,
                depth,
                geometry=geometry,
                use_numba=use_numba,
            ).propagator

    def propagator(self, depth):
        """Returns the propagator from the LUT which is closest to requested
        depth. If depth is outside the range of the propagators, function returns None.

        Parameters:
            depth     : float
                        refocus depth for requested propagator
        """

        # Find nearest propagator
        idx = self.closest_index(depth)
        if idx is not None:
            prop = Propagator(
                self.prop_table[idx, :, :],
                wavelength=self.wavelength,
                pixel_size=self.pixel_size,
                depth=self.depths[idx],
                geometry=self.geometry,
            )
            return prop
        else:
            return None

    """Returns the index of the propagator that is closest to requested
    depth. If depth is outside the range of the propagators, function returns None.

    Parameters:
        depth     : float
                    refocus depth for requested propagator
    """

    def closest_index(self, depth):
        if depth < self.depths[0] or depth > self.depths[-1]:
            return None

        elif self.num_depths == 1:  # Otherwise the algorithm to get the index will fail
            idx = 0

        else:
            idx = round(
                (depth - self.depths[0])
                / (self.depths[-1] - self.depths[0])
                * (self.num_depths - 1)
            )

        return idx

    def __str__(self):
        return (
            "LUT of "
            + str(self.num_depths)
            + " propagators from depth of "
            + str(self.depths[0])
            + " to "
            + str(self.depths[-1])
            + ". Wavelength: "
            + str(self.wavelength)
            + ", Pixel Size: "
            + str(self.pixel_size)
            + " ,Size:"
            + str(self.size)
        )



# -*- coding: utf-8 -*-
"""
PyHoloscope - Fast Holographic Microscopy for Python

Class to store stack of images numerically refocused to different depths.

"""

import math

import numpy as np
from PIL import Image


################### Class for stack of images focused at different depths ####
class FocusStack:
    def __init__(self, img, depth_range, num_depths):
        """Initialise instance of FocusStack class.

        Arguments:
            img        : numpy.ndarray
                         example image of the correct size and type, 2D numpy array
            num_depths : (float, float)
                         tuple of min depth and max depth in stack
            num_depths : int
                         number of images to be stored in stack
        """
        self.stack = np.zeros(
            (num_depths, np.shape(img)[0], np.shape(img)[1]), dtype=img.dtype
        )
        self.depths = np.linspace(depth_range[0], depth_range[1], num_depths)
        self.min_depth = depth_range[0]
        self.max_depth = depth_range[1]
        self.num_depths = num_depths
        self.depth_range = depth_range

    def __str__(self):
        return (
            "Refocus stack. Min: "
            + str(self.min_depth)
            + ", Max: "
            + str(self.max_depth)
            + ", Num: "
            + str(self.num_depths)
            + ", Step: "
            + str((self.max_depth - self.min_depth) / self.num_depths)
        )

    def add_idx(self, img, idx):
        """Adds an image to a specific index position.

        Arguments:
            img      : ndarray
                       image as a 2D numpy array
            idx      : int
                       index position
        """
        self.stack[idx, :, :] = img

    def add_depth(self, img, depth):
        """Adds an image to index position closest to the the specifed depth.

        Arguments:
            img      : ndarray
                       image as a 2D numpy array
            depth    : float
                       refocus depth of image
        """
        self.stack[self.depth_to_index(depth), :, :] = img

    def get_index(self, idx):
        """Returns the refocused image stored at the specified index.

        Arguments:
            idx      : int
                       index position to return image from

        Returns:
            numpy.ndarray : image

        """
        return self.stack[idx, :, :]

    def get_depth(self, depth):
        """Returns the closest refocused image to the specifed depth.

        Arguments:
            depth     : float
                        depth to return image from

        Returns:
            numpy.ndarray : image

        """
        return self.get_index(self.depth_to_index(depth))

    def get_depth_intensity(self, depth):
        """Returns the amplitude of the refocused image closest to the specified depth.

        Arguments:
            depth     : float
                        depth to return image from

        Returns:
            numpy.ndarray : image

        """
        return np.abs(self.get_depth(depth))

    def get_index_intensity(self, idx):
        """Return the amplitude of the refocused image at the specified index.

        Arguments:
            idx     : int
                      index position to return image from

        Returns:
            numpy.ndarray : image

        """
        return np.abs(self.get_index(idx))

    def depth_to_index(self, depth):
        """Returns the index closest to the specified depth.

        Arguments:
            depth     : float
                        depth to obtain closest index to

        Returns:
            int       : index

        """
        idx = round(
            (depth - self.min_depth)
            / (self.max_depth - self.min_depth)
            * (self.num_depths - 1)
        )
        if idx < 0:
            idx = 0
        if idx > self.num_depths - 1:
            idx = self.num_depths - 1
        return idx

    def index_to_depth(self, idx):
        """Returns depth corresponding to the specified index.

        Arguments:
            idx     : int
                      index position

        Returns:
            float   : depth

        """
        return self.depths[idx]

    def write_intensity_to_tif(self, filename, autoContrast=True):
        """Writes the amplitudes of the stack of refocused images to a 16 bit tif stack
        If autoContrast == True, all images will be autoscaled (across the whole
        stack, not individually) to use the full bit depth.

        Arguments:
            filename     : str
                           path/file to save to, should have .tif extension.

        Keyword Arguments:
            autoContrast : boolean
                           if True (default) images are autoscaled
        """

        if autoContrast:
            maxVal = np.max(np.abs(self.stack))
            minVal = np.min(np.abs(self.stack))

        imlist = []
        for m in self.stack:
            if autoContrast:
                im = np.abs(m).astype("float64") - minVal
                im = im / (maxVal - minVal) * 2**16
                imlist.append(Image.fromarray(im.astype("uint16")))

            else:
                imlist.append(Image.fromarray((255 * np.abs(m)).astype("uint16")))

        imlist[0].save(
            filename,
            compression="tiff_deflate",
            save_all=True,
            append_images=imlist[1:],
        )

    def write_phase_to_tif(self, filename):
        """Writes the phases of the stack of refocused images to a 16 bit tif stack.
        -pi is mapped to 0, pi is mapped to 255.

        Arguments:
            filename     : str
                           path/file to save to, should have .tif extension.
        """

        imlist = []
        for m in self.stack:
            im = (np.angle(m) + math.pi) * 255
            imlist.append(Image.fromarray(im.astype("uint16")))

        imlist[0].save(
            filename,
            compression="tiff_deflate",
            save_all=True,
            append_images=imlist[1:],
        )
